﻿

Add-Type -AssemblyName System.Drawing
Add-Type -AssemblyName System.Windows.Forms


$ID_OK = [System.Windows.Forms.MessageBoxButtons]::OK
$ID_YESNO = [System.Windows.Forms.MessageBoxButtons]::YesNo

$ICON_NONE = [System.Windows.Forms.MessageBoxIcon]::None
$ICON_INFO = [System.Windows.Forms.MessageBoxIcon]::Information
$ICON_ERROR = [System.Windows.Forms.MessageBoxIcon]::Error
$ICON_WARNING = [System.Windows.Forms.MessageBoxIcon]::Warning
$ICON_QUESTIONS = [System.Windows.Forms.MessageBoxIcon]::Question

##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################

[object] $script:Enabled         = $true
[bool]   $script:IgnoreOnlyDebug = $false
[object] $script:Window          = $null
[string] $script:TabСharacter    = "`t"


## Output-Enable
function Output-Enable
{
    Param
    (
        [Parameter(Position=0)]
        [bool] $Enable=$true
    )

    $script:Enabled = $Enable
}


## Output-IgnoreOnlyDebug
function Output-IgnoreOnlyDebug
{
    Param
    (
        [Parameter(Position=0)]
        [bool] $Ignore=$true
    )

    $script:IgnoreOnlyDebug = $Ignore    
}


## Output-SetWindow
function Output-SetWindow
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [object] $Window
    )

    $script:Window  = $Window
}


## Output-SetTabСharacter
function Output-SetTabСharacter
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [object] $TabСharacter
    )

    $script:TabСharacter  = $TabСharacter
}


## Output-Write
function Output-Write
{
    Param
    (
        [Parameter(Position = 0, Mandatory)]
        [Alias('Text')]
        [string]
        $Message,

        [Alias('Color')]
        [ConsoleColor[]]
        $ForegroundColor = 'White',

        [Alias('StartTab')]
        [ValidateRange(0, 10)]
        [int]
        $Indent = 0,

        [Alias('LinesBefore')]
        [ValidateRange(0, 10)]
        [int32]
        $LeadingSpace = 0,

        [Alias('LinesAfter')]
        [ValidateRange(0, 10)]
        [int32]
        $TrailingSpace = 0,

        [switch]
        $OnlyDebug,

        [switch]
        $ShowTime,
        
        [switch]
        $NoNewLine
    )
  
    if (-not $script:Enabled) {return}

    # Add leading lines
    Write-Host ("`n" * $LeadingSpace) -NoNewline

    # Add TABS before text
    Write-Host ($script:TabСharacter * $Indent) -NoNewline

    Write-Host -ForegroundColor $ForegroundColor $Message -NoNewline

    if (-not $NoNewLine) {
        Write-Host
    }
    
    # Add leading lines
    Write-Host ("`n" * $TrailingSpace) -NoNewline

    if ((-not $OnlyDebug -or $script:IgnoreOnlyDebug) -and ($script:Window -ne $null))
    {
        # Add leading lines
        for ($i = 0; $i -lt $LeadingSpace; $i++) {
            $script:Window.AppendText([Environment]::NewLine)
        }

        $script:Window.AppendText($script:TabСharacter * $Indent)
        $script:Window.AppendText($Message)


        if (-not $NoNewLine) 
        {
            $script:Window.AppendText([Environment]::NewLine)
        }

        # Add leading lines
        for ($i = 0; $i -lt $TrailingSpace; $i++) {
            $script:Window.AppendText([Environment]::NewLine)
        }

        $script:Window.SelectionStart = $script:Window.Text.Length
        $script:Window.ScrollToCaret()
    }

    #   $msg | Out-File "./export.txt" -Append; # <<< or add a param to location
}


## Output-Result
function Output-Result
{
    Param
    (
        [Parameter(Position = 0, Mandatory)]        
        [bool]
        $Result,

        [Alias('Color')]
        [ConsoleColor[]]
        $ForegroundColor = 'White',

        [Alias('StartTab')]
        [ValidateRange(0, 10)]
        [int]
        $Indent = 0,

        [Alias('LinesBefore')]
        [ValidateRange(0, 10)]
        [int32]
        $LeadingSpace = 0,

        [Alias('LinesAfter')]
        [ValidateRange(0, 10)]
        [int32]
        $TrailingSpace = 0,

        [switch]
        $OnlyDebug        
    )    

    $Message         = $(if ($result) {"- Успех"} else {"■ Неудача"})
    $ForegroundColor = $(if ($result) {"White"}   else {"Red"})

    if ($OnlyDebug)
    {
        Output-Write -Message $Message -ForegroundColor $ForegroundColor -Indent $Indent -LinesBefore $LeadingSpace -LinesAfter $TrailingSpace -OnlyDebug
    }
    else
    {
        Output-Write -Message $Message -ForegroundColor $ForegroundColor -Indent $Indent -LinesBefore $LeadingSpace -LinesAfter $TrailingSpace
    }
}


## Output-Warning
function Output-Warning
{
    Param
    (
        [Parameter(Position = 0, Mandatory)]
        [Alias('Text')]
        [string]
        $Message,

        [ValidateRange(0, 10)]
        [int]
        $Indent = 0,

        [switch]
        $OnlyDebug
    )    

    if ($OnlyDebug)
    {
        Output-Write -Message "◾ Предупреждение: $Message" -Indent $Indent -OnlyDebug
    }
    else
    {        
        Output-Write -Message "◾ Предупреждение: $Message" -Indent $Indent
    }
}


## Output-Clear
function Output-Clear
{
    Clear-Host

    if ($script:Window -ne $null)
    {
        $script:Window.ResetText()
    }
}


##########################################################################

function Write-Color 
{
    [CmdletBinding()]
    param(
        [Parameter(Position = 0, Mandatory, ValueFromPipeline)]
        [Alias('Text')]
        [string[]]
        $Message,

        [Alias('Color', 'FGC')]
        [ConsoleColor[]]
        $ForegroundColor = 'White',

        [Alias('BGC')]
        [ConsoleColor[]]
        $BackgroundColor,

        [Alias('StartTab')]
        [ValidateRange(0, 5)]
        [int]
        $Indent = 0,

        [Alias('LinesBefore')]
        [ValidateRange(0, 10)]
        [int]
        $LeadingSpace = 0,

        [Alias('LinesAfter')]
        [ValidateRange(0, 10)]
        [int]
        $TrailingSpace = 0,

        [Alias('File', 'L')]
        [ValidateScript( { Test-Path -IsValid $_ })]
        [string]
        $LogFile,

        [Alias('DateFormat', 'TimeFormat')]
        [string]
        $DateTimeFormat = "yyyy-MM-dd HH:mm:ss",

        [Alias('LTS')]
        [switch]
        $NoLogTimestamp,

        [Alias('Encoding')]
        [System.Text.Encoding]
        $OutputEncoding = [System.Text.Encoding]::UTF8,

        [switch]
        $ShowTime,

        [switch]
        $NoNewLine
    )
    begin {
        if ($PSBoundParameters.ContainsKey('LogFile')) { $StringBuilder = [System.Text.StringBuilder]::new() }

        $InvalidParams =    ($BackgroundColor.Count -gt 0 -and $BackgroundColor.Count -ne $ForegroundColor.Count) -or
                            $ForegroundColor.Count -gt $Message.Count -or $BackgroundColor.Count -gt $Message.Count

        if ($InvalidParams) {
            $ErrorRecord = [System.Management.Automation.ErrorRecord]::new(
                [System.Management.Automation.ParameterBindingException]::new(
                    'The number of provided Foreground and Background colors does not match, or more colors were provided than entered messages.'
                ),
                'InvalidColorMapping', # ErrorID
                [System.Management.Automation.ErrorCategory]::InvalidArgument, # Category
                $null # Target
            )

            $PSCmdlet.ThrowTerminatingError($ErrorRecord)
        }

        # Add leading lines
        Write-Host ("`n" * $LeadingSpace) -NoNewline

        if ($ShowTime) {
            # Add Time before output
            Write-Host ('[{0}]' -f [datetime]::Now.ToString($DateTimeFormat)) @BaseParams
        }
    }
    process {
        # Add TABS before text
        Write-Host ("`t" * $Indent) -NoNewline

        if ($PSBoundParameters.ContainsKey('ForegroundColor') -or $PSBoundParameters.ContainsKey('BackgroundColor')) {
            # The real deal coloring
            for ($i = 0; $i -lt $Message.Count; $i++) {

                if ($BackgroundColor[$i] -ne [console]::BackgroundColor){
                    $WriteParams = @{
                        NoNewLine       = $true
                        ForegroundColor = $ForegroundColor[$i]
                        BackgroundColor = $BackgroundColor[$i]
                    }
                }
                else{
                    $WriteParams = @{
                        NoNewLine       = $true
                        ForegroundColor = $ForegroundColor[$i]
                    }
                } 

                Write-Host $Message[$i] @WriteParams

                if ($PSBoundParameters.ContainsKey('LogFile')) { $StringBuilder.Append($Message[$i]) > $null }
            }
        } 
        else {
            Write-Host $Message -NoNewline
        }

        if (-not $NoNewLine) {
            Write-Host
        }

        # Add empty line after
        Write-Host ("`n" * $TrailingSpace) -NoNewline
    }
    end {
        if ($PSBoundParameters.ContainsKey('LogFile')) {
            # Save to file

            try {
                if ($NoLogTimestamp) {
                    $StringBuilder.ToString() | Add-Content -Path $LogFile -Encoding $OutputEncoding
                } else {
                    "[{0}]{1}" -f [datetime]::Now.ToString($DateTimeFormat), $StringBuilder.ToString() |
                        Add-Content -Path $LogFile -Encoding $OutputEncoding
                }
            } catch {
                $PSCmdlet.WriteError($_)
            }
        }
    }
}



##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################


##########################################################################
## module data
[string] $script:ComputerName = ""
[string] $script:LoggedInUser = ""

##########################################################################
## Get computer name
function UserMgt_GetComputerName
{
    Param
    (           
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Get computer name" -Indent $StackLevel -OnlyDebug  
    
    if ($script:ComputerName -eq "")
    {
        Output-Write "- Undefined. Getting the computer name..." -Indent $StackLevel -OnlyDebug 

        $local:SystemInfo    = Get-WmiObject -Class Win32_ComputerSystem
        $script:ComputerName = $local:SystemInfo.DNSHostName 
    }

    $local:result = if ($script:ComputerName -ne "") {$script:ComputerName} else {$null}
    Output-Write "result: $local:result " -Indent $StackLevel -OnlyDebug
    return $local:result
}


##########################################################################
## Get the name of the logged in user
function UserMgt_GetLoggedIn
{
    Param
    (           
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Get the name of the logged in user" -Indent $StackLevel -OnlyDebug  
    
    if ($script:LoggedInUser -eq "")
    {
        Output-Write "- Undefined. Getting the user name..." -Indent $StackLevel -OnlyDebug 
        
        $local:wmi = $(Get-WMIObject -class Win32_ComputerSystem | select username)
        if ($local:wmi -eq $null)
        {
            Output-Write " ◾ failed to create Win32_ComputerSystem object" -Indent $StackLevel -OnlyDebug
            Output-Write "result: null" -OnlyDebug
            return $null
        }

        $script:LoggedInUser = $local:wmi.username
        if ($script:LoggedInUser -eq "")
        {
            Output-Write " ◾ couldn't get the username of the logged in user" -Indent $StackLevel -OnlyDebug
            Output-Write "result: null" -Indent $StackLevel -OnlyDebug
            return $null
        }

        $script:LoggedInUser = $script:LoggedInUser.Split("\")[-1]
    }

    $local:result = if ($script:LoggedInUser -ne "") {$script:LoggedInUser} else {$null}
    Output-Write "result: $local:result " -Indent $StackLevel -OnlyDebug
    return $local:result
}


##########################################################################
## Get the SID of an object by its name
function UserMgt_GetSID
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Name')]
        [string] $ObjectName,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Get object SID" -Indent $StackLevel -OnlyDebug  
    Output-Write " - Name:     `t$ObjectName" -Indent $StackLevel -OnlyDebug

    try 
    {
        $local:id  = new-object System.Security.Principal.NTAccount($ObjectName)
        $local:sid = $local:id.Translate( [System.Security.Principal.SecurityIdentifier] ).toString()        

        Output-Write "result: $local:sid" -Indent $StackLevel -OnlyDebug
        return $local:sid        
    }
    catch 
    {        
    }

    Output-Write " ◾ couldn't get the object's sid" -Indent $StackLevel -OnlyDebug
    Output-Write "result: null" -Indent $StackLevel -OnlyDebug
    return $null
}


##########################################################################
## Get the name of an object by its SID
function UserMgt_GetName
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('SID')]
        [string] $ObjectSID,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Get object Name" -Indent $StackLevel -OnlyDebug
    Output-Write " - SID:     `t$ObjectSID" -Indent $StackLevel -OnlyDebug

    try 
    {        
        $local:si     = New-Object System.Security.Principal.SecurityIdentifier($ObjectSID)
	    $local:nt     = $local:si.Translate( [System.Security.Principal.NTAccount])
	    $local:name   = $local:nt.ToString()	
	    $local:name   = $local:name.split('\')[-1] -replace '\.\w+$'

        Output-Write "result: $local:name" -Indent $StackLevel -OnlyDebug
        return $local:name
    }
    catch 
    {
    }

    Output-Write " ◾ couldn't get the object's name" -Indent $StackLevel -OnlyDebug
    Output-Write "result: null" -Indent $StackLevel -OnlyDebug
    return $null
}


##########################################################################
## Check the existence of the user or group by name
function UserMgt_IsObjectExist
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Name')]
        [string] $ObjectName,        

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Check the existence of the user or group by name" -Indent $StackLevel -OnlyDebug
    Output-Write " - Name:     `t$ObjectName" -Indent $StackLevel -OnlyDebug

    try 
    {
        $local:ComputerName = $(UserMgt_GetComputerName $($StackLevel+1))
        if ($local:ComputerName -eq $null)
        {
            Output-Write " ◾ couldn't get the computer name" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        if ([adsi]::Exists("WinNT://$($local:ComputerName)/$ObjectName"))
        {
            Output-Write "result: true" -Indent $StackLevel -OnlyDebug
            return $true
        }
    }
    catch 
    {
        Output-Write " ◾ failed" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    Output-Write " ◾ user or group not found" -Indent $StackLevel -OnlyDebug
    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}


##########################################################################
# Check the presence of the user or group by SID
function UserMgt_IsObjectExistBySID
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('SID')]
        [string] $ObjectSID,        

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Check the existence of the user or group by sid" -Indent $StackLevel -OnlyDebug
    Output-Write " - SID:     `t$ObjectSID" -Indent $StackLevel -OnlyDebug

    try
    {
        $local:si = New-Object System.Security.Principal.SecurityIdentifier($ObjectSID)
	    $null     = $local:si.Translate( [System.Security.Principal.NTAccount])
        
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }
    catch 
    {
    } 

    Output-Write " ◾ user or group not found" -Indent $StackLevel -OnlyDebug
    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}


##########################################################################
## Validate username
function UserMgt_IsUserameValid
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Name')]
        [string] $UserName,        

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Validate username" -Indent $StackLevel -OnlyDebug
    Output-Write " - Name:     `t$UserName" -Indent $StackLevel -OnlyDebug

    if (($UserName.Length -eq 0) -or ($UserName.Length -gt 20))
    {
        Output-Write " ◾ invalid username length" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    if (($UserName -Replace "\.| |`t").Length -eq 0)
    {
        Output-Write " ◾ user name cannot consist only of spaces and dots" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    if ($UserName.LastIndexOfAny("/\`"*[]:|<>+=;,?@") -ne -1)
    {
        Output-Write " ◾ user name contains one or more prohibited characters: /\`"*[]:|<>+=;,?@" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    Output-Write "result: true" -Indent $StackLevel -OnlyDebug
    return $true
}


##########################################################################
## Check user credentials
function UserMgt_IsCredintalsValid
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Name')]
        [string] $UserName,
        
        [Parameter(Position=1, Mandatory)]
        [Alias('Password')]
        [SecureString] $SecurePassword,

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    try 
    { 
        $local:BSTR = [System.Runtime.InteropServices.Marshal]::SecureStringToBSTR($SecurePassword)
        $local:Password = [System.Runtime.InteropServices.Marshal]::PtrToStringAuto($local:BSTR)
    }
    catch
    {
        Output-Write " ◾ invalid param" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    Output-Write "Check user credentials" -Indent $StackLevel -OnlyDebug
    Output-Write " - Name:     `t$UserName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Password: `t$("*" * $local:Password.Length)" -Indent $StackLevel -OnlyDebug

    if ($(UserMgt_IsObjectExist $UserName $($StackLevel+1)) -eq $false)
    {
        Output-Write " ◾ user not found" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    try
    {
        $local:ComputerName = $(UserMgt_GetComputerName $($StackLevel+1))
        if ($local:ComputerName -eq $null)
        {
            Output-Write " ◾ couldn't get the computer name" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        Add-Type -AssemblyName System.DirectoryServices.AccountManagement
        $local:obj = New-Object System.DirectoryServices.AccountManagement.PrincipalContext('machine', $($local:ComputerName))

        if ($local:obj.ValidateCredentials($UserName, $local:Password))
        {
            Output-Write "result: true" -Indent $StackLevel -OnlyDebug
            return $true
        }
                
        Output-Write " ◾ invalid password" -Indent $StackLevel -OnlyDebug
    }
    catch
    {
        Output-Write " ◾ invalid credentials" -Indent $StackLevel -OnlyDebug
    }

    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}


##########################################################################
## Create new user
function UserMgt_CreateUser
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Name')]
        [string] $UserName,
        
        [Parameter(Position=1, Mandatory)]
        [Alias('Password')]
        [SecureString] $SecurePassword,

        [Parameter(Position=2, Mandatory=$false)]
        [string] $FullName,

        [Parameter(Position=3, Mandatory=$false)]
        [string] $Description,

        [Parameter(Position=4, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    try 
    { 
        $local:BSTR = [System.Runtime.InteropServices.Marshal]::SecureStringToBSTR($SecurePassword)
        $local:Password = [System.Runtime.InteropServices.Marshal]::PtrToStringAuto($local:BSTR)
    }
    catch
    {
        Output-Write " ◾ invalid param" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    Output-Write "Create a new user" -Indent $StackLevel -OnlyDebug
    Output-Write " - Name:        `t$UserName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Password:    `t$("*" * $local:Password.Length)" -Indent $StackLevel -OnlyDebug
    if ($FullName -ne "")    {Output-Write " - FullName:    `t$FullName" -Indent $StackLevel -OnlyDebug}
    if ($Description -ne "") {Output-Write " - Description: `t$Description" -Indent $StackLevel -OnlyDebug}


    if ($(UserMgt_IsObjectExist $UserName $($StackLevel+1)))
    {

        if ($(UserMgt_IsCredintalsValid $UserName $SecurePassword $($StackLevel+1)))
        {
            Output-Write " -  user already exists, credentials are correct" -Indent $StackLevel -OnlyDebug
            Output-Write "result: true" -Indent $StackLevel -OnlyDebug
            return $true  
        }
        else
        {
            Output-Write " ◾ user already exists, password is incorrect" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }
    }

    try
    {
        $local:ComputerName = $(UserMgt_GetComputerName $($StackLevel+1))
        if ($local:ComputerName -eq $null)
        {
            Output-Write " ◾ couldn't get the computer name" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        try
        {
            $local:Computer = [adsi]"WinNT://$($local:ComputerName), computer"
            $local:User = $local:Computer.Create('user', $UserName)
        }
        catch
        {
            Output-Write " ◾ failed to create object 'user'" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false 
        }

	    $null = $local:User.SetPassword($local:Password)	 
        $null = $local:User.Put('description', $Description)
        $null = $local:User.Put('fullname', $FullName)
	    $null = $local:User.SetInfo()

        # password never expires
	    $local:User.UserFlags.Value = $local:User.UserFlags.Value -bor 0x10040

        $null = $local:User.CommitChanges()
    }
    catch 
    {
        Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    Output-Write "result: ok" -Indent $StackLevel -OnlyDebug
    return $true
}


##########################################################################
## Check if the user is a member of the group
function UserMgt_IsMemberOfGroup
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('User')]
        [string] $UserName,

        [Parameter(Position=1, Mandatory)]
        [Alias('Group')]
        [string] $GroupName,

        [Parameter(Position=4, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Check if the user is a member of the group" -Indent $StackLevel -OnlyDebug
    Output-Write " - User:     `t$UserName"  -Indent $StackLevel -OnlyDebug
    Output-Write " - Group:    `t$GroupName" -Indent $StackLevel -OnlyDebug

    if (-not $(UserMgt_IsObjectExist $UserName $($StackLevel+1)))
    {
        Output-Write " ◾ the user does not exist" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    if (-not $(UserMgt_IsObjectExist $GroupName $($StackLevel+1)))
    {
        Output-Write " ◾ the group does not exist" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }    

    try 
    {
        $local:ComputerName = $(UserMgt_GetComputerName $($StackLevel+1))
        if ($local:ComputerName -eq $null)
        {
            Output-Write " ◾ couldn't get the computer name" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        try
        {
            $local:Group = [ADSI]"WinNT://$($local:ComputerName)/$GroupName, group"
        }
        catch
        {
            Output-Write " ◾ failed to create object 'group'" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false 
        }

        try
        {
            $local:User  = [ADSI]"WinNT://$($local:ComputerName)/$UserName,  user"
        }
        catch
        {
            Output-Write " ◾ failed to create object 'user'" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false 
        }
        
        if ($local:Group.IsMember($local:User.ADsPath))
        {
            Output-Write "result: true" -Indent $StackLevel -OnlyDebug
            return $true    
        }
    }
    catch 
    {
        Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    Output-Write " ◾ the user is not a member of the group" -Indent $StackLevel -OnlyDebug
    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}


##########################################################################
#Add user to group
function UserMgt_AddToGroup
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('User')]
        [string] $UserName,

        [Parameter(Position=1, Mandatory)]
        [Alias('Group')]
        [string] $GroupName,

        [Parameter(Position=4, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Check if the user is a member of the group" -Indent $StackLevel -OnlyDebug
    Output-Write " - User:     `t$UserName"  -Indent $StackLevel -OnlyDebug
    Output-Write " - Group:    `t$GroupName" -Indent $StackLevel -OnlyDebug

    if ($(UserMgt_IsMemberOfGroup $UserName $GroupName $($StackLevel+1)))
    {
        Output-Write " -  the user is already a member of the group" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }

    try
    {
        $local:ComputerName = $(UserMgt_GetComputerName $($StackLevel+1))
        if ($local:ComputerName -eq $null)
        {
            Output-Write " ◾ couldn't get the computer name" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        try
        {
            $local:Group = [ADSI]"WinNT://$($local:ComputerName)/$GroupName, group"
        }
        catch
        {
            Output-Write " ◾ failed to create object 'group'" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false 
        }

        try
        {
            $local:User  = [ADSI]"WinNT://$($local:ComputerName)/$UserName,  user"
        }
        catch
        {
            Output-Write " ◾ failed to create object 'user'" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false 
        }

        $null = $local:Group.Add($local:User.ADsPath)
    }
    catch 
    {
        Output-Write " ◾ failed to add user to group" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $false
    }

    Output-Write "result: ok" -Indent $StackLevel -OnlyDebug
    return $true
}

##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################



##########################################################################
## Reset DCOM settings to defaults
function DCOM_ResetSettingsToDefault
{
    Param
    (
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Reset DCOM settings to defaults..." -Indent $StackLevel -OnlyDebug

    try
    {
        $local:COMAdmin = New-Object -com ("COMAdmin.COMAdminCatalog.1")
        $local:LocalCollections = $local:COMAdmin.Connect("localhost")
        $local:LocalComputer    = $local:LocalCollections.GetCollection("LocalComputer", $local:LocalCollections.Name)
        $local:LocalComputer.Populate()

        $local:LocalComputerItem = $local:LocalComputer.Item(0)

        Output-Write " - Setting this computer's Transaction Timeout to 60" -Indent $StackLevel -OnlyDebug
        $local:LocalComputerItem.Value("TransactionTimeout") = 60                
    
        Output-Write " - Enabling DCOM on this computer" -Indent $StackLevel -OnlyDebug
        $local:LocalComputerItem.Value("DCOMEnabled") = $true

        #Output-Write " - Enabling COM Internet Services on this computer..."
        #$LocalComputerItem.Value("CISEnabled") = $false

        Output-Write " - Setting COM default authentication level to Connect" -Indent $StackLevel -OnlyDebug
        $local:LocalComputerItem.Value("DefaultAuthenticationLevel") = 2        

        Output-Write " - Setting COM default impersonation level to Impersonate" -Indent $StackLevel -OnlyDebug
        $local:LocalComputerItem.Value("DefaultImpersonationLevel") = 3            

        Output-Write " - Saving changes" -Indent $StackLevel -OnlyDebug
        $local:LocalComputer.SaveChanges()
    }
    catch
    {
        Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $true
    }

    Output-Write "result: true" -Indent $StackLevel -OnlyDebug
    return $true
}


##########################################################################
##Updating the SDDL value in the registry key
function DCOM_UpdateSDDL
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Path')]
        [string] $RegistryPath,

        [Parameter(Position=1, Mandatory)]
        [Alias('Name')]
        [string] $PropertyName,

        [Parameter(Position=2, Mandatory)]
        [Alias('SID')]
        [string] $ObjectSID,

        [Parameter(Position=3, Mandatory)]
        [Alias('Value')]
        [string] $Permission,

        [Parameter(Position=4, Mandatory)]
        [Alias('Default')]
        [string] $DefaultSDDL,

        [Parameter(Position=5, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    $local:AddedSDDL = "$Permission$ObjectSID"

    Output-Write "Updating the SDDL value in the registry key..." -Indent $StackLevel -OnlyDebug
    Output-Write " - Path:      `t$RegistryPath" -Indent $StackLevel -OnlyDebug
    Output-Write " - Property:  `t$PropertyName" -Indent $StackLevel -OnlyDebug
    Output-Write " - AddedSDDL: `t$AddedSDDL"    -Indent $StackLevel -OnlyDebug
    Output-Write " - Default:   `t$DefaultSDDL"  -Indent $StackLevel -OnlyDebug

    $local:Converter = new-object system.management.ManagementClass Win32_SecurityDescriptorHelper 

    try
    {
        Get-Item -Path $RegistryPath  -ErrorAction Stop | Out-Null
        Get-ItemProperty -Path $RegistryPath | Select-Object -ExpandProperty $PropertyName -ErrorAction Stop | Out-Null

        $local:RegistryItem = Get-ItemProperty -Path $RegistryPath -Name $PropertyName
        Output-Write "- The registry key was found" -Indent $StackLevel -OnlyDebug

        if ($local:RegistryItem.$PropertyName.Length -eq 0)
        {
            Output-Write " - Invalid registry key value detected (1)" -Indent $StackLevel -OnlyDebug
            Remove-ItemProperty –Path $RegistryPath –Name $PropertyName
            Output-Write " - Registry key removed" -Indent $StackLevel -OnlyDebug
            throw 
        }

        $local:CurrentSDDL = $local:Converter.BinarySDToSDDL($local:RegistryItem.$PropertyName)
        if ($local:CurrentSDDL.SDDL.Length -eq 0)
        {
            Output-Write " - Invalid registry key value detected (2)" -Indent $StackLevel -OnlyDebug
            Remove-ItemProperty –Path $RegistryPath –Name $PropertyName
            Output-Write " - Registry key removed" -Indent $StackLevel -OnlyDebug
            throw 
        }

        $local:CurrentBinarySD = $local:Converter.SDDLToBinarySD($local:CurrentSDDL)
        Output-Write " - Value:  `t$($local:CurrentSDDL.SDDL)" -Indent $StackLevel -OnlyDebug
    }
    catch
    {
        Output-Write "- Registry key not found" -Indent $StackLevel -OnlyDebug

        $local:CurrentBinarySD = $local:Converter.SDDLToBinarySD($local:DefaultSDDL)
        $local:CurrentSDDL     = $local:Converter.BinarySDToSDDL($local:CurrentBinarySD.BinarySD)
                
        New-ItemProperty –Path $RegistryPath –Name $PropertyName -PropertyType "Binary"
        Output-Write " - Registry key created" -Indent $StackLevel -OnlyDebug
        
        Set-ItemProperty -Path $RegistryPath -Name $PropertyName -Value $local:CurrentBinarySD.BinarySD
        Output-Write " - The default value is set" -Indent $StackLevel -OnlyDebug
    }

    # PartialMatch
    $local:PartialMatchSDDL = "\(\w;;\w+;;;$ObjectSID\)"
    
    # Building new permissions
    Output-Write "- Building new permissions..." -Indent $StackLevel -OnlyDebug
    if ($local:CurrentSDDL.SDDL -match $local:PartialMatchSDDL)
    {
        Output-Write " - Partial permission match found, required changes were made" -Indent $StackLevel -OnlyDebug
        $local:CurrentSDDL.SDDL = $local:CurrentSDDL.SDDL -replace $local:PartialMatchSDDL, ""
    }
    else
    {
        Output-Write " - No permissions found, required changes were made" -Indent $StackLevel -OnlyDebug
    }
    $local:NewSDDL = $local:CurrentSDDL.SDDL + "(" + $local:AddedSDDL + ")"
    Output-Write " - Value:  `t$($local:NewSDDL)" -Indent $StackLevel -OnlyDebug

    # Convert SDDL back to Binary
    $local:NewBinarySD = $local:Converter.SDDLToBinarySD($local:NewSDDL)

    if ($local:NewBinarySD.BinarySD.Length -eq 0)
    {
        Output-Write " ◾ convert SDDL to Binary failed" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    # Apply the changes...
    Set-ItemProperty -Path $RegistryPath -Name $PropertyName -Value $local:NewBinarySD.BinarySD
    Output-Write " - Changes saved in the registry key." -Indent $StackLevel -OnlyDebug

    Output-Write "result: true" -Indent $StackLevel -OnlyDebug

    return $true
}


##########################################################################
## Add user or group (by SID) to 'Defaults' DCOM security settings
function DCOM_AddToDefaultsBySID
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('SID')]
        [string] $ObjectSID,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Add user or group (by SID) to 'Defaults' DCOM security settings..." -Indent $StackLevel -OnlyDebug  
    Output-Write " - SID:     `t$ObjectSID" -Indent $StackLevel -OnlyDebug        

    if (-not $(UserMgt_IsObjectExistBySID $ObjectSID $($StackLevel+1)))
    {
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    $local:OlePath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Ole"

    # AccessPermission - Local Access, Remote Access
    $local:AccessPermission  = "A;;CCDCLC;;;"
    $local:DefaultAccessSDDL = "O:BAG:BAD:(A;;CCDCLC;;;PS)(A;;CCDC;;;SY)(A;;CCDCLC;;;BA)"
    $local:result1 = DCOM_UpdateSDDL -Path $local:OlePath -Name "DefaultAccessPermission" -SID $ObjectSID -Value $local:AccessPermission -Default $local:DefaultAccessSDDL -Stack $($StackLevel+1)

    # LaunchPermission - Local Launch, Remote Launch, Local Activation, Remote Activation
    $local:LaunchPermission  = "A;;CCDCLCSWRP;;;"
    $local:DefaultLaunchSDDL = "O:BAG:BAD:(A;;CCDCLCSWRP;;;BA)(A;;CCDCSW;;;IU)(A;;CCDCSW;;;SY)"
    $local:result2 = DCOM_UpdateSDDL -Path $local:OlePath -Name "DefaultLaunchPermission" -SID $ObjectSID -Value $local:LaunchPermission -Default $local:DefaultLaunchSDDL -Stack $($StackLevel+1)
    
    $result = $($result1 -and $result2)
    if (-not $result)
    {
        Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
    }
    Output-Write "result: $result" -Indent $StackLevel -OnlyDebug
    return $result
}


##########################################################################
## Add user or group to 'Defaults' DCOM security settings
function DCOM_AddToDefaults
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Name')]
        [string] $ObjectName,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Add user or group to 'Defaults' DCOM security settings..." -Indent $StackLevel -OnlyDebug  
    Output-Write " - Name:     `t$ObjectName" -Indent $StackLevel -OnlyDebug            

    $local:ObjectSID = UserMgt_GetSID $ObjectName $($StackLevel+1)
    if ($local:ObjectSID -eq $null)
    {
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    if ($(DCOM_AddToDefaultsBySID $ObjectSID $($StackLevel+1)))
    {
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }

    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}


##########################################################################
## Check if the DCOM parameters need manual repair
function DCOM_IsLimitsValid
{
    Param
    (
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Check if the DCOM parameters need manual repair..." -Indent $StackLevel -OnlyDebug    
    
    $local:OlePath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Ole"

    [bool] $result = $true
    try
    {
        Get-Item -Path $local:OlePath  -ErrorAction Stop | Out-Null
        Get-ItemProperty -Path $local:OlePath | Select-Object -ExpandProperty "MachineAccessRestriction" -ErrorAction Stop | Out-Null
    }
    catch
    {
        Output-Write " ◾ registry parameter '$local:OlePath\MachineAccessRestriction' not found" -Indent $StackLevel -OnlyDebug
        $result = $result -band $false
    }

    try
    {
        Get-Item -Path $local:OlePath  -ErrorAction Stop | Out-Null
        Get-ItemProperty -Path $local:OlePath | Select-Object -ExpandProperty "MachineLaunchRestriction" -ErrorAction Stop | Out-Null
    }
    catch
    {
        Output-Write " ◾ registry parameter '$local:OlePath\MachineLaunchRestriction' not found" -Indent $StackLevel -OnlyDebug
        $result = $result -band $false
    }
    
    Output-Write "result: $result" -Indent $StackLevel -OnlyDebug
    return $result
}


##########################################################################
## Add user or group (by SID) to 'Limits' DCOM security settings
function DCOM_AddToLimitsBySID
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('SID')]
        [string] $ObjectSID,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Add user or group (by SID) to 'Limits' DCOM security settings..." -Indent $StackLevel -OnlyDebug  
    Output-Write " - SID:     `t$ObjectSID" -Indent $StackLevel -OnlyDebug        

    if (-not $(DCOM_IsLimitsValid $($StackLevel+1)))
    {
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    if (-not $(UserMgt_IsObjectExistBySID $ObjectSID $($StackLevel+1)))
    {
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    $local:OlePath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Ole"

    # AccessPermission - Local Access, Remote Access
    $local:AccessPermission  = "A;;CCDCLC;;;"
    $local:DefaultAccessSDDL = "O:BAG:BAD:(A;;CCDCLC;;;PS)(A;;CCDC;;;SY)(A;;CCDCLC;;;BA)"
    $local:result1 = DCOM_UpdateSDDL -Path $local:OlePath -Name "MachineAccessRestriction" -SID $ObjectSID -Value $local:AccessPermission -Default $local:DefaultAccessSDDL -Stack $($StackLevel+1)
#    Start-Sleep -Milliseconds 5000

    # LaunchPermission - Local Launch, Remote Launch, Local Activation, Remote Activation
    $local:LaunchPermission  = "A;;CCDCLCSWRP;;;"
    $local:DefaultLaunchSDDL = "O:BAG:BAD:(A;;CCDCLCSWRP;;;BA)(A;;CCDCSW;;;IU)(A;;CCDCSW;;;SY)"
    $local:result2 = DCOM_UpdateSDDL -Path $local:OlePath -Name "MachineLaunchRestriction" -SID $ObjectSID -Value $local:LaunchPermission -Default $local:DefaultLaunchSDDL -Stack $($StackLevel+1)
#    Start-Sleep -Milliseconds 5000

    $result = $($result1 -and $result2)
    if (-not $result)
    {
        Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
    }
    Output-Write "result: $result" -Indent $StackLevel -OnlyDebug
    return $result
}


##########################################################################
## Add user or group to 'Limits' DCOM security settings
function DCOM_AddToLimits
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Name')]
        [string] $ObjectName,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Add user or group to 'Limits' DCOM security settings..." -Indent $StackLevel -OnlyDebug  
    Output-Write " - Name:     `t$ObjectName" -Indent $StackLevel -OnlyDebug            

    $local:ObjectSID = UserMgt_GetSID $ObjectName $($StackLevel+1)
    if ($local:ObjectSID -eq $null)
    {
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    if ($(DCOM_AddToLimitsBySID $ObjectSID $($StackLevel+1)))
    {
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }

    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}


##########################################################################
## Сheck if the DСOM application is installed
function DCOM_IsAppExist
{
    Param
    (
        [Parameter(Position=0, Mandatory)]        
        [string] $AppID,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Сheck if the DСOM application is installed..." -Indent $StackLevel -OnlyDebug
    Output-Write " - AppID:     `t$AppID" -Indent $StackLevel -OnlyDebug

    $local:AppPath = "Registry::HKEY_CLASSES_ROOT\AppID\{$AppID}"

    try
    {            
        Get-Item -Path $local:AppPath  -ErrorAction Stop | Out-Null

        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }
    catch
    {
        Output-Write " ◾ could not find DCOM application" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

}


##########################################################################
## Get the name of DCOM application
function DCOM_GetAppName
{
    Param
    (
        [Parameter(Position=0, Mandatory)]        
        [string] $AppID,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Get the name of DCOM application..." -Indent $StackLevel -OnlyDebug
    Output-Write " - AppID:     `t$AppID" -Indent $StackLevel -OnlyDebug
    
    $local:wmi = Get-WMIObject -Class Win32_DCOMApplicationSetting -Filter "AppID='{$AppID}'" -EnableAllPrivileges

    if ($local:wmi -eq $null)
    {
        Output-Write " ◾ could not find DCOM application" -Indent $StackLevel -OnlyDebug
        Output-Write "result: null" -Indent $StackLevel -OnlyDebug
        return $null
    }
    else
    {        
        Output-Write "result: $($wmi.Caption)" -Indent $StackLevel -OnlyDebug
        return $wmi.Caption
    }
}


##########################################################################
## Get the service name corresponding to the DCOM application
function DCOM_GetAppServiceName
{
    Param
    (
        [Parameter(Position=0, Mandatory)]        
        [string] $AppID,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Get the service name corresponding to the DCOM application..." -Indent $StackLevel -OnlyDebug
    Output-Write " - AppID:     `t$AppID" -Indent $StackLevel -OnlyDebug
    
    $local:wmi = Get-WMIObject -Class Win32_DCOMApplicationSetting -Filter "AppID='{$AppID}'" -EnableAllPrivileges

    if ($local:wmi -eq $null)
    {
        Output-Write " ◾ could not find DCOM application" -Indent $StackLevel -OnlyDebug
        Output-Write "result: null" -Indent $StackLevel -OnlyDebug
        return $null
    }
    else
    {        
        Output-Write "result: $($wmi.LocalService)" -Indent $StackLevel -OnlyDebug
        return $wmi.LocalService
    }
}


##########################################################################
## Get the path to the service executable file corresponding to the DCOM application
function DCOM_GetAppServicePath
{
    Param
    (
        [Parameter(Position=0, Mandatory)]        
        [string] $AppID,

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Get the path to the service executable file corresponding to the DCOM application..." -Indent $StackLevel -OnlyDebug
    Output-Write " - AppID:     `t$AppID" -Indent $StackLevel -OnlyDebug
    
    $local:ServiceName = $(DCOM_GetAppServiceName $AppID $($StackLevel+1))
    if ($local:ServiceName -eq $null)
    {
        Output-Write "result: null" -Indent $StackLevel -OnlyDebug
        return $null
    }

    $local:Service = get-wmiobject -query "select * from win32_service where name='$local:ServiceName'"
    if ($local:Service -eq $null)
    {
        Output-Write " ◾ could not find service" -Indent $StackLevel -OnlyDebug
        Output-Write "result: null" -Indent $StackLevel -OnlyDebug
        return $null
    }

    $local:expIdx = $($local:Service.pathname).IndexOf(".exe")
    if ($local:expIdx -lt 0)
    {
        Output-Write " ◾ could not get the path to the service executable" -Indent $StackLevel -OnlyDebug
        Output-Write "result: null" -Indent $StackLevel -OnlyDebug
        return $null     
    }

    $local:ServicePath = $($local:Service.pathname).substring(0, 1).tolower() + $($local:Service.pathname).substring(1, $expIdx+(4-1))
    Output-Write "result: $($local:ServicePath)" -Indent $StackLevel -OnlyDebug
    return $local:ServicePath
}


# Check if the application is using the default value for the parameter
function DCOM_IsAppUsesDefaultValue
{
    Param
    (
        [Parameter(Position=0, Mandatory)]        
        [string] $AppID,

        [Parameter(Position=1, Mandatory)]
        [Alias('Name')]
        [string] $PropertyName,

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Check if the application is using the default value for the parameter..." -Indent $StackLevel -OnlyDebug
    Output-Write " - AppID:     `t$AppID" -Indent $StackLevel -OnlyDebug
    Output-Write " - Property:  `t$PropertyName" -Indent $StackLevel -OnlyDebug

    $local:AppPath = "Registry::HKEY_CLASSES_ROOT\AppID\{$AppID}"

    try
    {            
        Get-Item -Path $local:AppPath  -ErrorAction Stop | Out-Null
    }
    catch
    {
        Output-Write " ◾ could not find DCOM application" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }


    try
    {
        Get-ItemProperty -Path $local:AppPath | Select-Object -ExpandProperty $PropertyName -ErrorAction Stop | Out-Null

        Output-Write " ◾ the application uses a non-default value for the parameter" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }
    catch
    {
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }
}


function DCOM_UpdateAppPermissions
{
    Param
    (
        [Parameter(Position=0, Mandatory)]        
        [string] $AppID,

        [Parameter(Position=1, Mandatory)]
        [string] $ObjectName,

        [Parameter(Position=3, Mandatory)]
        [string] $Permission,

        [Parameter(Position=4, Mandatory)]        
        [string[]] $Rights,

        [Parameter(Position=5, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Update user or group permissions in DCOM application permissions..." -Indent $StackLevel -OnlyDebug
    Output-Write " - AppID:     `t$AppID"      -Indent $StackLevel -OnlyDebug
    Output-Write " - ObjectName:`t$ObjectName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Permission:`t$Permission" -Indent $StackLevel -OnlyDebug
    Output-Write " - Rights:    `t$Rights"     -Indent $StackLevel -OnlyDebug

    [int32] $local:AccesMask = 1
    foreach ($Right in $Rights)
    {
        switch ($Right)
        {
            "Local Launch"      {$local:AccesMask = $AccesMask -bOR 2}
            "Remote Launch"     {$local:AccesMask = $AccesMask -bOR 4}
            "Local Activation"  {$local:AccesMask = $AccesMask -bOR 8}
            "Remote Activation" {$local:AccesMask = $AccesMask -bOR 16}

            "Local Access"      {$local:AccesMask = $AccesMask -bOR 27}
            "Remote Access"     {$local:AccesMask = $AccesMask -bOR 29}
        }
    }
    
    try
    {
        $local:wmi = Get-WMIObject -Class Win32_DCOMApplicationSetting -Filter "AppID='{$AppID}'" -EnableAllPrivileges
        if ($local:wmi -eq $null)
        {
            Output-Write " ◾ could not find DCOM application" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        if (-not $(UserMgt_IsObjectExist $ObjectName $($StackLevel+1)))
        {
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        $local:CurrentSD = $null
        Switch ($Permission)
        {
            "LaunchAndActivationPermission" {$local:CurrentSD  = $local:wmi.GetLaunchSecurityDescriptor().descriptor}
            "AccessPermission"              {$local:CurrentSD  = $local:wmi.GetAccessSecurityDescriptor().descriptor}
            "ConfigurationPermission"       {$local:CurrentSD  = $local:wmi.GetConfigurationSecurityDescriptor().descriptor}
        }

        if ($local:CurrentSD -eq $null)
        {
            Output-Write " ◾ invalid value for the 'Permission' parameter" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        # Create empty Win32_SecurityDescriptor object
        $local:NewSD = $([wmiclass]”\\.\root\cimv2:Win32_SecurityDescriptor”).CreateInstance()
        if ($local:NewSD -eq $null)
        {
            Output-Write " ◾ failed to create security descriptor" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        $local:NewSD.ControlFlags = $local:CurrentSD.ControlFlags
        $local:NewSD.Group = $local:CurrentSD.Group
        $local:NewSD.Owner = $local:CurrentSD.Owner
        $local:NewSD.SACL  = $local:CurrentSD.SACL
        $local:NewSD.Owner = $local:CurrentSD.Owner

        # Remove user if it already has permissions
        ForEach ($local:oace in $local:CurrentSD.DACL)
        {
            if ($ObjectName -in $local:oace.Trustee.Name)
            { 
                Output-Write "- Remove user or group from the list of DACL's" -Indent $StackLevel -OnlyDebug
            }
            else
            {
                $local:NewSD.DACL += $local:oace
            }
        }

        # Add user or group (again) with correct permissions
        Output-Write "- Add user or group to the list of DACL's" -Indent $StackLevel -OnlyDebug
        $local:Trustee = ([wmiclass]'Win32_Trustee').psbase.CreateInstance()
        $local:Trustee.Name   = $ObjectName

        $local:ace = ([wmiclass]'Win32_ACE').psbase.CreateInstance()
        $local:ace.AccessMask = $local:AccesMask
        $local:ace.Trustee    = $local:Trustee        

        $NewSD.DACL += [System.Management.ManagementBaseObject]$Ace

        Switch ($Permission)
        {
            "LaunchAndActivationPermission" {$null = $wmi.SetLaunchSecurityDescriptor($local:NewSD)}
            "AccessPermission"              {$null = $wmi.SetAccessSecurityDescriptor($local:NewSD)}
            "ConfigurationPermission"       {$null = $wmi.SetAccessSecurityDescriptor($local:NewSD)}
        }

    }
    catch
    {
        Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }
       

    Output-Write "result: ok" -Indent $StackLevel -OnlyDebug
    return $true
}


##########################################################################
## Add permission to the user to use the DСOM application
function DCOM_AddUserPermissionsToApp
{
    Param
    (
        [Parameter(Position=0, Mandatory)]        
        [string] $AppID,

        [Parameter(Position=1, Mandatory)]
        [string] $ObjectName,

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Add permission to the user or group to use the DСOM application..." -Indent $StackLevel -OnlyDebug
    Output-Write " - AppID:     `t$AppID" -Indent $StackLevel -OnlyDebug
    Output-Write " - ObjectName:`t$ObjectName" -Indent $StackLevel -OnlyDebug

    $local:ComputerName = $(UserMgt_GetComputerName $($StackLevel+1))
    if ($local:ComputerName -eq $null)
    {
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    if (-not $(UserMgt_IsObjectExist $ObjectName $($StackLevel+1)))
    {
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    $local:App = Get-WMIObject -class Win32_DCOMApplication -Filter "AppID='{$AppID}'" -ComputerName $local:ComputerName
    if($local:App -eq $null)
	{
        Output-Write " ◾ could not find DCOM application" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
	}  

    try
    {
        $local:wmi = Get-WMIObject -Class Win32_DCOMApplicationSetting -Filter "AppID='{$AppID}'" -EnableAllPrivileges
        if ($local:wmi -eq $null)
        {
            Output-Write " ◾ could not find DCOM application" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        $local:wmi.AuthenticationLevel = 2
        $null = $local:wmi.put()
    }
    catch
    {
        Output-Write "result: ◾ failed to set the AuthenticationLevel value" -Indent $StackLevel -OnlyDebug
        return $false
    }

    $result1 = $(DCOM_ResetSettingsToDefault $($StackLevel+1))
    $result2 = $(DCOM_AddToDefaults $ObjectName $($StackLevel+1))
    $result3 = $(DCOM_AddToLimits   $ObjectName $($StackLevel+1))

    $result4 = $true    
                 
    if (-not $(DCOM_IsAppUsesDefaultValue $AppID "LaunchPermission" $($StackLevel+1)))
    {
        $result4 = $(DCOM_UpdateAppPermissions $AppID $ObjectName "LaunchAndActivationPermission" @("Local Launch", "Remote Launch", "Local Activation", "Remote Activation") $($StackLevel+1))
    }

    $result5 = $true

    if (-not $(DCOM_IsAppUsesDefaultValue $AppID "AccessPermission" $($StackLevel+1)))
    {
        $result5 = $(DCOM_UpdateAppPermissions $AppID $ObjectName "AccessPermission" @("Local Access", "Remote Access") $($StackLevel+1))
    }


    $result = $($result1 -and $result2 -and $result3 -and $result4 -and $result5)
    if (-not $result)
    {
        Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
    }
    Output-Write "result: $result" -Indent $StackLevel -OnlyDebug
    return $result
}


##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################


##########################################################################
## Grant the rights to control the service
function SvcUtil_GrantServiceControlRights
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Service')]
        [string] $ServiceName,

        [Parameter(Position=1, Mandatory)]
        [Alias('SID')]
        [string] $ObjectSID,

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Grant the rights to control the service" -Indent $StackLevel -OnlyDebug
    Output-Write " - Service: `t$ServiceName" -Indent $StackLevel -OnlyDebug
    Output-Write " - SID:     `t$ObjectSID"   -Indent $StackLevel -OnlyDebug        


    if (-not $(UserMgt_IsObjectExistBySID $ObjectSID $($StackLevel+1)))
    {
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    $local:CurrentSDDL = (& $env:systemroot\system32\sc sdshow $ServiceName)
    if ($local:CurrentSDDL.Count -ne 2)
    {
        Output-Write " ◾ specified service is not installed" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    $local:CurrentSDDL =  $local:CurrentSDDL[1]
    Output-Write " - SDDL:  `t$($local:CurrentSDDL)" -Indent $StackLevel -OnlyDebug

    # PartialMatch
    $local:PartialMatchSDDL = "\(\w;;\w+;;;$ObjectSID\)"
    
    # Building new permissions
    Output-Write "- Building new permissions..." -Indent $StackLevel -OnlyDebug
    if ($local:CurrentSDDL -match $PartialMatchSDDL)
    {
        Output-Write " - Partial permission match found, required changes were made" -Indent $StackLevel -OnlyDebug
        $local:CurrentSDDL = $local:CurrentSDDL -replace $local:PartialMatchSDDL, ""
    }
    else
    {
        Output-Write " - No permissions found, required changes were made" -Indent $StackLevel -OnlyDebug
    }

    $local:FullControlRights = "CCDCLCSWRPWPDTLOCRSDRCWDWO"
    $local:AddedSDDL = "A;;$local:FullControlRights;;;$ObjectSID"

    $local:NewSDDL = $local:CurrentSDDL.Replace("D:(", "D:($local:AddedSDDL)(")
    Output-Write " - SDDL:  `t$($local:NewSDDL)" -Indent $StackLevel -OnlyDebug


    (& $env:systemroot\system32\sc sdset $ServiceName $local:NewSDDL) | out-null

    Output-Write "result: true" -Indent $StackLevel -OnlyDebug
    return $true
}


##########################################################################
# Grant the rights to control the service by AppID
function SvcUtil_GrantServiceControlRightsByAppID
{
    Param
    (
        [Parameter(Position=0, Mandatory)]        
        [string] $AppID,

        [Parameter(Position=1, Mandatory)]
        [Alias('SID')]
        [string] $ObjectSID,

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Grant the rights to control the service" -Indent $StackLevel -OnlyDebug
    Output-Write " - AppID:   `t$AppID"     -Indent $StackLevel -OnlyDebug
    Output-Write " - SID:     `t$ObjectSID" -Indent $StackLevel -OnlyDebug

    $local:ServiceName = $(DCOM_GetAppServiceName $AppID $($StackLevel+1))    
    if ($local:ServiceName -eq $null)
    {
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }


    $result = $(SvcUtil_GrantServiceControlRights $local:ServiceName $ObjectSID $($StackLevel+1))    

    Output-Write "result: $result" -Indent $StackLevel -OnlyDebug
    return $result        
}


##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################


##########################################################################
# Grant the user access rights to the folder 
function FS_GrantAccessRightsToFolder
{
    Param
    (
        [Parameter(Position=0, Mandatory)]        
        [string] $Path,

        [Parameter(Position=1, Mandatory)]
        [Alias('User')]
        [string] $UserName,

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Grant the user access rights to the folder " -Indent $StackLevel -OnlyDebug
    Output-Write " - Path:     `t$Path"     -Indent $StackLevel -OnlyDebug
    Output-Write " - User:     `t$UserName" -Indent $StackLevel -OnlyDebug

    if ($(UserMgt_IsObjectExist $UserName $($StackLevel+1)) -eq $false)
    {
        Output-Write " ◾ user not found" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    try
    {
        $local:ComputerName = $(UserMgt_GetComputerName $($StackLevel+1))

        $local:user = "$local:ComputerName\$UserName"
        $local:acl = Get-Acl -Path $Path

        # disable inheritance for a folder
        $acl.SetAccessRuleProtection($True, $True)
        Set-Acl -Path $Path -AclObject $local:acl

        # remove folder access rights
        $local:acl        = Get-Acl -Path $Path
        $local:rules      = $local:acl.Access | where IsInherited -eq $false
        $local:targetrule = $rules | where IdentityReference -eq $local:user
        if ($targetrule -ne $null)
        {
            if ($targetrule.Count -eq 0)
            {
                $null = $local:acl.RemoveAccessRule($local:targetrule)
                $local:acl | Set-Acl -Path $Path
            }
            else
            {
                for ($i = 0; $i -lt $local:targetrule.Count; $i++) 
                {        
                    $null = $local:acl.RemoveAccessRule($local:targetrule[$i])
                    $local:acl | Set-Acl -Path $Path
                }
            }
        }

        # add folder access rights
        $local:acl = Get-Acl $Path
        $local:rule = New-Object -TypeName System.Security.AccessControl.FileSystemAccessRule -ArgumentList $local:user, "Fullcontrol", "Containerinherit, ObjectInherit", "None", "Allow"
        $null = $local:acl.SetAccessRule($local:rule)
        $local:acl | Set-Acl -Path $Path

        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }
    catch
    {        
        Output-Write " ◾ the access rights to the `"$Path`" directory are not configured correctly" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }
}


##########################################################################
## module data
[object] $script:lsp_data = $null


##########################################################################
## Export of local security policy settings
function LSP_Export
{
    Param
    (
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Export of local security policy settings" -Indent $StackLevel -OnlyDebug

    if ($script:lsp_data -ne $null)
    {
        Output-Write " ◾ the local security policy settings have already been exported" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    try
    {
        $local:tmp = [System.IO.Path]::GetTempFileName()
#$null = MessageBox "secedit 1." $main_form.Text $ID_OK $ICON_ERROR        
#$local:tmp = "d:\1.inf"
        secedit /export /cfg "$local:tmp.inf" | Out-Null
#$null = MessageBox "secedit 2." $main_form.Text $ID_OK $ICON_ERROR        
        $script:lsp_data = New-Object PSObject
                
        # parse *.inf
        $local:contents = Get-Content "$local:tmp.inf" -raw
        $local:index = 0
        [regex]::Matches($local:contents,"(?<=\[)(.*)(?=\])") | %{
            $local:title = $_
            [regex]::Matches($local:contents,"(?<=\]).*?((?=\[)|(\Z))", [System.Text.RegularExpressions.RegexOptions]::Singleline)[$local:index] | %{
                $local:section = new-object psobject
                $_.value -split "\r\n" | ?{$_.length -gt 0} | %{
                    $local:value = [regex]::Match($_,"(?<=\=).*").value
                    $local:name = [regex]::Match($_,".*(?=\=)").value
                    $local:section | Add-Member -MemberType NoteProperty -Name $local:name.tostring().trim() -Value $local:value.tostring().trim() -ErrorAction SilentlyContinue | out-null
                }
                $script:lsp_data | Add-Member -MemberType NoteProperty -Name $local:title -Value $local:section | out-null
            }
            $local:index += 1
        }
    
        rm "$local:tmp.inf" -ea 0 | out-null

        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }
    catch
    {
        Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }
}

##########################################################################
## Confirming changes to local security policy
function LSP_Rollback
{
    $script:lsp_data = $null
}


##########################################################################
## Confirming changes to local security policy
function LSP_CommitChanges
{
    Param
    (
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Confirming changes to local security policy" -Indent $StackLevel -OnlyDebug

    if ($script:lsp_data -eq $null)
    {
        Output-Write " ◾ unable to confirm changes, local security policy settings were not exported" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    $local:tmp = [System.IO.Path]::GetTempFileName()
#$local:tmp = "d:\2.inf"
    try
    {
        $script:lsp_data.PSObject.Properties.GetEnumerator() | %{
            "[$($_.Name)]"
            $_.Value | %{
                $_.psobject.Properties.GetEnumerator() | %{
                    "$($_.Name) = $($_.Value)"}
            }
        } | out-file "$local:tmp.inf" -ErrorAction Stop

        $script:lsp_data = $null

        secedit /import /cfg "$local:tmp.inf" /db "$tmp.sdb"    | out-null
        secedit /configure /db "$local:tmp.sdb" /cfg "$tmp.inf" | out-null

        rm "$local:tmp.inf" -ea 0 | out-null
        rm "$local:tmp.sdb" -ea 0 | out-null

        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }
    catch
    {
        Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }   
}


##########################################################################
## Is there a value in the local security policy
function LSP_IsValueExist
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Section')]
        [String] $SectionName,
        
        [Parameter(Position=1, Mandatory)]
        [Alias('Param')]
        [String] $ParamName,

        [Parameter(Position=2, Mandatory)]
        [Alias('Value')]
        [String] $ParamValue,

        [Parameter(Position=3, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Is there a value in the local security policy" -Indent $StackLevel -OnlyDebug
    Output-Write " - Section:     `t$SectionName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Param:       `t$ParamName"   -Indent $StackLevel -OnlyDebug
    Output-Write " - Value:       `t$ParamValue"  -Indent $StackLevel -OnlyDebug

    if ($script:lsp_data -eq $null)
    {
        Output-Write " ◾ local security policy settings were not exported" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    if ($script:lsp_data.$SectionName -eq $null)
    {
        Output-Write " ◾ section does not exist" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }   

    if ($script:lsp_data.$SectionName.$ParamName -eq $null)
    {
        Output-Write " ◾ parameter does not exist" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    $local:Values = $script:lsp_data.$SectionName.$ParamName.Split(",")
    if ($local:Values -contains $ParamValue.Trim())
    {            
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }

    Output-Write "◾ value does not exist" -Indent $StackLevel -OnlyDebug
    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}


##########################################################################
## Change the value of the local security policy setting
function LSP_ChangeValue
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Section')]
        [String] $SectionName,
        
        [Parameter(Position=1, Mandatory)]
        [Alias('Param')]
        [String] $ParamName,

        [Parameter(Position=2, Mandatory)]
        [Alias('Value')]
        [String] $ParamValue,

        [Parameter(Position=3, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Change the value of the local security policy setting" -Indent $StackLevel -OnlyDebug
    Output-Write " - Section:     `t$SectionName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Param:       `t$ParamName"   -Indent $StackLevel -OnlyDebug
    Output-Write " - Value:       `t$ParamValue"  -Indent $StackLevel -OnlyDebug

    if ($script:lsp_data.$SectionName -eq $null)
    {
        Output-Write " ◾ section does not exist" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }   

    if ($script:lsp_data.$SectionName.$ParamName -eq $null)
    {
        $script:lsp_data.$SectionName | add-member -MemberType NoteProperty -Name $ParamName -Value $ParamValue.Trim()

        Output-Write " -  parameter not found and was created" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }

    if ($script:lsp_data.$SectionName.$ParamName -eq $ParamValue)
    {
        Output-Write " -  parameter does not need to be changed" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }

    $script:lsp_data.$SectionName.$ParamName = $ParamValue

    Output-Write "result: true" -Indent $StackLevel -OnlyDebug    
    return $true
}


##########################################################################
## Add value to local security policy setting
function LSP_AddValue
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Section')]
        [String] $SectionName,
        
        [Parameter(Position=1, Mandatory)]
        [Alias('Param')]
        [String] $ParamName,

        [Parameter(Position=2, Mandatory)]
        [Alias('Value')]
        [String] $ParamValue,

        [Parameter(Position=3, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Add value to local security policy setting" -Indent $StackLevel -OnlyDebug
    Output-Write " - Section:     `t$SectionName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Param:       `t$ParamName"   -Indent $StackLevel -OnlyDebug
    Output-Write " - Value:       `t$ParamValue"  -Indent $StackLevel -OnlyDebug

    if ($script:lsp_data.$SectionName -eq $null)
    {
        Output-Write " ◾ section does not exist" -Indent $StackLevel -OnlyDebug
        Output-Write "result: false" -Indent $StackLevel -OnlyDebug
        return $false
    }

    if ($script:lsp_data.$SectionName.$ParamName -eq $null)
    {
        $script:lsp_data.$SectionName | add-member -MemberType NoteProperty -Name $ParamName -Value $ParamValue.Trim()

        Output-Write " -  parameter not found and was created" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }

    if ($script:lsp_data.$SectionName.$ParamName -eq "")
    {
        $script:lsp_data.$SectionName.$ParamName = $ParamValue

        Output-Write " -  the parameter existed, but the value was empty" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug        
        return $true
    }

    $local:Values = $script:lsp_data.$SectionName.$ParamName.Split(",")
    if ($local:Values -contains $ParamValue.Trim())
    {
        Output-Write " -  value already exists" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }
    else
    {
        $local:Values += $ParamValue.Trim()
        $script:lsp_data.$SectionName.$ParamName = $local:Values -join ','
    }    

    Output-Write "result: true" -Indent $StackLevel -OnlyDebug    
    return $true
}


##########################################################################
## Remove the value from the local security policy setting
function LSP_RemoveValue
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Section')]
        [String] $SectionName,
        
        [Parameter(Position=1, Mandatory)]
        [Alias('Param')]
        [String] $ParamName,

        [Parameter(Position=2, Mandatory)]
        [Alias('Value')]
        [String] $ParamValue,

        [Parameter(Position=3, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Remove the value from the local security policy setting" -Indent $StackLevel -OnlyDebug
    Output-Write " - Section:     `t$SectionName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Param:       `t$ParamName"   -Indent $StackLevel -OnlyDebug
    Output-Write " - Value:       `t$ParamValue"  -Indent $StackLevel -OnlyDebug

    if ($script:lsp_data.$SectionName -eq $null)
    {
        Output-Write " -  section does not exist" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }

    if ($script:lsp_data.$SectionName.$ParamName -eq $null)
    {
        Output-Write " -  parameter does not exist" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }

    $local:Values = $script:lsp_data.$SectionName.$ParamName.Split(",")
    if (-not ($local:Values -contains $ParamValue.Trim()))
    {
        Output-Write " -  the value does not exist" -Indent $StackLevel -OnlyDebug
        Output-Write "result: true" -Indent $StackLevel -OnlyDebug
        return $true
    }
    else
    {
        $local:Values = $local:Values | Where-Object { $_ -ne $ParamValue.Trim() }

        if ($local:Values.Count -eq 0)
        {
            $script:lsp_data.$SectionName.$ParamName = ""
        }
        else
        {
            $script:lsp_data.$SectionName.$ParamName = $local:Values -join ','
        }            
    }

    Output-Write "result: true" -Indent $StackLevel -OnlyDebug    
    return $true
}


##########################################################################
## Assigning object Rights
function LSP_AssigningObjectRights
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Section')]
        [String] $SectionName,
        
        [Parameter(Position=1, Mandatory)]
        [Alias('Param')]
        [String] $ParamName,

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('User')]
        [String] $UserName = "",

        [Parameter(Position=3, Mandatory=$false)]
        [Alias('Group')]
        [String] $GroupSID = "",

        [Parameter(Position=4, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Assigning user or group Rights" -Indent $StackLevel -OnlyDebug
    Output-Write " - Section:     `t$SectionName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Param:       `t$ParamName"   -Indent $StackLevel -OnlyDebug
    if ($UserName -ne "") {Output-Write " - User:        `t$UserName"  -Indent $StackLevel -OnlyDebug}
    if ($GroupSID -ne "") {Output-Write " - Group:       `t$GroupSID"  -Indent $StackLevel -OnlyDebug}

    if ($UserName -ne "")
    {
        if (-not $(UserMgt_IsObjectExist $UserName $($StackLevel+1)))
        {        
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

         $result = $(LSP_AddValue $SectionName $ParamName $UserName $($StackLevel+1))
    
        Output-Write "result: $result" -Indent $StackLevel -OnlyDebug
        return $result
    }
    elseif ($GroupSID -ne "")
    {
        if (-not $(UserMgt_IsObjectExistBySID $GroupSID $($StackLevel+1)))
        {        
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        $result = $(LSP_AddValue $SectionName $ParamName "*$GroupSID" $($StackLevel+1))
    
        Output-Write "result: $result" -Indent $StackLevel -OnlyDebug
        return $result
    }

    Output-Write " ◾ invalid param" -Indent $StackLevel -OnlyDebug
    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}


##########################################################################
## Revoke object Rights
function LSP_RevokeObjectRights
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Section')]
        [String] $SectionName,
        
        [Parameter(Position=1, Mandatory)]
        [Alias('Param')]
        [String] $ParamName,

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('User')]
        [String] $UserName = "",

        [Parameter(Position=3, Mandatory=$false)]
        [Alias('Group')]
        [String] $GroupSID = "",

        [Parameter(Position=4, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Revoke user or group Rights" -Indent $StackLevel -OnlyDebug
    Output-Write " - Section:     `t$SectionName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Param:       `t$ParamName"   -Indent $StackLevel -OnlyDebug
    if ($UserName -ne "") {Output-Write " - User:        `t$UserName"  -Indent $StackLevel -OnlyDebug}
    if ($GroupSID -ne "") {Output-Write " - Group:       `t$GroupSID"  -Indent $StackLevel -OnlyDebug}


    if ($UserName -ne "")
    {
        if (-not $(UserMgt_IsObjectExist $UserName $($StackLevel+1)))
        {        
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

         $result = $(LSP_RemoveValue $SectionName $ParamName $UserName $($StackLevel+1))
    
        Output-Write "result: $result" -Indent $StackLevel -OnlyDebug
        return $result
    }
    elseif ($GroupSID -ne "")
    {
        if (-not $(UserMgt_IsObjectExistBySID $GroupSID $($StackLevel+1)))
        {        
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        $result = $(LSP_RemoveValue $SectionName $ParamName "*$GroupSID" $($StackLevel+1))
    
        Output-Write "result: $result" -Indent $StackLevel -OnlyDebug
        return $result
    }

    Output-Write " ◾ invalid param" -Indent $StackLevel -OnlyDebug
    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}


##########################################################################
## Grant the rights to Log on as a service
function LSP_GrantServiceLogonRight
{
    Param
    (
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('User')]
        [String] $UserName = "",

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Group')]
        [String] $GroupSID = "",

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Grant the rights to Log on as a service" -Indent $StackLevel -OnlyDebug
    if ($UserName -ne "") {Output-Write " - User:        `t$UserName"  -Indent $StackLevel -OnlyDebug}
    if ($GroupSID -ne "") {Output-Write " - Group:       `t$GroupSID"  -Indent $StackLevel -OnlyDebug}

    $local:Params = @{
        SectionName = "Privilege Rights"
        ParamName   = "SeServiceLogonRight" 
        StackLevel  = $StackLevel+1
    }

    if ($UserName -ne "") {$local:Params += @{UserName = $UserName}}
    if ($GroupSID -ne "") {$local:Params += @{GroupSID = $GroupSID}}

    $local:result = LSP_AssigningObjectRights @local:Params

    Output-Write "result: $local:result" -Indent $StackLevel -OnlyDebug
    return $local:result
}


##########################################################################
## Deny Log on Locally
function LSP_DenyLogonLocally
{
    Param
    (
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('User')]
        [String] $UserName = "",

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Group')]
        [String] $GroupSID = "",

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Deny user or group to log in locally" -Indent $StackLevel -OnlyDebug
    if ($UserName -ne "") {Output-Write " - User:        `t$UserName"  -Indent $StackLevel -OnlyDebug}
    if ($GroupSID -ne "") {Output-Write " - Group:       `t$GroupSID"  -Indent $StackLevel -OnlyDebug}


    $local:AssigningParams = @{
        SectionName = "Privilege Rights"
        ParamName   = "SeDenyInteractiveLogonRight" 
        StackLevel  = $StackLevel+1
    }

    $local:RevokeParams = @{
        SectionName = "Privilege Rights"
        ParamName   = "SeInteractiveLogonRight" 
        StackLevel  = $StackLevel+1
    }

    if ($UserName -ne "") 
    {
        if ($UserName.Split("\")[-1] -eq $(UserMgt_GetLoggedIn $($StackLevel+1)))
        {
            Output-Write " ◾ you cannot deny logon locally for a logged in user. this can cause the system to crash after a restart" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        $local:AssigningParams += @{UserName = $UserName}
        $local:RevokeParams    += @{UserName = $UserName}
    }
    elseif ($GroupSID -ne "") 
    {
        $local:ObjectName = $(UserMgt_GetName $GroupSID $($StackLevel+1))
        if ($local:ObjectName.Split("\")[-1] -eq $(UserMgt_GetLoggedIn $($StackLevel+1)))
        {
            Output-Write " ◾ you cannot deny logon locally for a logged in user. this can cause the system to crash after a restart" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }

        $local:AssigningParams += @{GroupSID = $GroupSID}
        $local:RevokeParams    += @{GroupSID = $GroupSID}
    }

    $local:result1 = LSP_AssigningObjectRights @local:AssigningParams
    $local:result2 = LSP_RevokeObjectRights @local:RevokeParams

    Output-Write "result: $($local:result1 -and $local:result2)" -Indent $StackLevel -OnlyDebug
    return $($local:result1 -and $local:result2)
}


##########################################################################
## Deny access to this computer from the network
function LSP_DenyAccessFromNetwork
{
    Param
    (
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('User')]
        [String] $UserName = "",

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Group')]
        [String] $GroupSID = "",

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )
    
    Output-Write "Deny access to this computer from the network" -Indent $StackLevel -OnlyDebug
    if ($UserName -ne "") {Output-Write " - User:        `t$UserName"  -Indent $StackLevel -OnlyDebug}
    if ($GroupSID -ne "") {Output-Write " - Group:       `t$GroupSID"  -Indent $StackLevel -OnlyDebug}


    $local:AssigningParams = @{
        SectionName = "Privilege Rights"
        ParamName   = "SeDenyNetworkLogonRight" 
        StackLevel  = $StackLevel+1
    }

    $local:RevokeParams = @{
        SectionName = "Privilege Rights"
        ParamName   = "SeNetworkLogonRight" 
        StackLevel  = $StackLevel+1
    }

    if ($UserName -ne "") 
    {
        $local:AssigningParams += @{UserName = $UserName}
        $local:RevokeParams    += @{UserName = $UserName}
    }
    elseif ($GroupSID -ne "") 
    {
        $local:AssigningParams += @{GroupSID = $GroupSID}
        $local:RevokeParams    += @{GroupSID = $GroupSID}
    }

    $local:result1 = LSP_AssigningObjectRights @local:AssigningParams
    $local:result2 = LSP_RevokeObjectRights @local:RevokeParams

    Output-Write "result: $($local:result1 -and $local:result2)" -Indent $StackLevel -OnlyDebug
    return $($local:result1 -and $local:result2)
}


##########################################################################
## Allow access to this computer from the network
function LSP_AllowAccessFromNetwork
{
    Param
    (
        [Parameter(Position=0, Mandatory=$false)]
        [Alias('User')]
        [String] $UserName = "",

        [Parameter(Position=1, Mandatory=$false)]
        [Alias('Group')]
        [String] $GroupSID = "",

        [Parameter(Position=2, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )
    
    Output-Write "Allow access to this computer from the network" -Indent $StackLevel -OnlyDebug
    if ($UserName -ne "") {Output-Write " - User:        `t$UserName"  -Indent $StackLevel -OnlyDebug}
    if ($GroupSID -ne "") {Output-Write " - Group:       `t$GroupSID"  -Indent $StackLevel -OnlyDebug}


    $local:AssigningParams = @{
        SectionName = "Privilege Rights"
        ParamName   = "SeNetworkLogonRight"
        StackLevel  = $StackLevel+1
    }

    $local:RevokeParams = @{
        SectionName = "Privilege Rights"
        ParamName   = "SeDenyNetworkLogonRight"
        StackLevel  = $StackLevel+1
    }

    if ($UserName -ne "") 
    {
        $local:AssigningParams += @{UserName = $UserName}
        $local:RevokeParams    += @{UserName = $UserName}
    }
    elseif ($GroupSID -ne "") 
    {
        $local:AssigningParams += @{GroupSID = $GroupSID}
        $local:RevokeParams    += @{GroupSID = $GroupSID}
    }

    $local:result1 = LSP_AssigningObjectRights @local:AssigningParams
    $local:result2 = LSP_RevokeObjectRights @local:RevokeParams

    Output-Write "result: $($local:result1 -and $local:result2)" -Indent $StackLevel -OnlyDebug
    return $($local:result1 -and $local:result2)
}



##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################




function Update-NetFirewallRule
{
    Param
    (
        [Parameter(Position=0, Mandatory)]
        [Alias('Name')]
        [string] $DisplayName,
        
        [Parameter(Position=1, Mandatory)] 
        [object] $Direction,
        
        [Parameter(Position=2, Mandatory)]
        [object] $Action,
                
        [Parameter(Position=3, Mandatory=$false)]
        [String] $Protocol = "",

        [Parameter(Position=4, Mandatory=$false)] 
        [string] $LocalPort = "",

        [Parameter(Position=5, Mandatory=$false)]
        [String] $Program = "",
        
        [Parameter(Position=6, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Create a firewall rule or update it if the rule already exists" -Indent $StackLevel -OnlyDebug
    Output-Write " - Name:        `t$DisplayName" -Indent $StackLevel -OnlyDebug
    Output-Write " - Direction:   `t$Direction"   -Indent $StackLevel -OnlyDebug
    Output-Write " - Action:      `t$Action"  -Indent $StackLevel -OnlyDebug
    if ($Protocol -ne "")  {Output-Write " - Protocol:    `t$Protocol"  -Indent $StackLevel -OnlyDebug}
    if ($LocalPort -ne "") {Output-Write " - LocalPort:   `t$LocalPort"  -Indent $StackLevel -OnlyDebug}
    if ($Program -ne "")   {Output-Write " - Program:     `t$Program"  -Indent $StackLevel -OnlyDebug}


    $RuleParams = @{
        DisplayName = $DisplayName
        Direction   = $Direction 
        Action      = $Action 
        Enabled     = "True"
    }

    if ($Protocol -ne "")   {$RuleParams += @{Protocol  = $Protocol}}
    if ($LocalPort -ne "")  {$RuleParams += @{LocalPort = $LocalPort}}
    if ($Program -ne "")    {$RuleParams += @{Program   = $Program}}

    Output-Write "- RuleParams: $($RuleParams | ConvertTo-Json -Compress)" -Indent $StackLevel -OnlyDebug

    try 
    {
        $null = Get-NetFirewallRule -DisplayName $DisplayName -ErrorAction Stop

        try
        {
            Output-Write "- firewall rule already exists, it will be updated" -Indent $StackLevel -OnlyDebug

            $null = Set-NetFirewallRule @RuleParams

            Output-Write "result: ok" -Indent $StackLevel -OnlyDebug
            return $true
        }
        catch
        {
            Output-Write " ◾ failed to update a firewall rule" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }        
    }
    catch
    {
        try
        {
            Output-Write "- firewall rule does not exist, a new rule will be created" -Indent $StackLevel -OnlyDebug

            $null = New-NetFirewallRule @RuleParams

            Output-Write "result: ok" -Indent $StackLevel -OnlyDebug
            return $true
        }
        catch
        {
            Output-Write " ◾ failed to create a firewall rule" -Indent $StackLevel -OnlyDebug
            Output-Write "result: false" -Indent $StackLevel -OnlyDebug
            return $false
        }
    }

    
    Output-Write " ◾ unknown error" -Indent $StackLevel -OnlyDebug
    Output-Write "result: false" -Indent $StackLevel -OnlyDebug
    return $false
}





##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################
##########################################################################

function MessageBox
{
    Param
    (
        [Parameter(Position=0, Mandatory)] 
        [string] $Text,
        
        [Parameter(Position=1, Mandatory)] 
        [string] $Title,
        
        [Parameter(Position=2, Mandatory=$false)] 
        [string] $Buttons=$ID_OK,
        
        [Parameter(Position=3, Mandatory=$false)]
        [string] $Icon=$ICON_NONE
    )

    return [System.Windows.Forms.MessageBox]::Show($Text, $Title, $Buttons, $Icon) 
}


function ShellExecute 
{
    Param
    (
        [Parameter(Position=0, Mandatory)] 
        [string] $Path,
        
        [Parameter(Position=1, Mandatory)] 
        [string] $ArgList,
        
        [Parameter(Position=2, Mandatory=$false)] 
        [string] $Wait = $true,

        [Parameter(Position=6, Mandatory=$false)]
        [Alias('Stack')]
        [int32] $StackLevel = 0
    )

    Output-Write "Start process" -Indent $StackLevel -OnlyDebug
    Output-Write " - Path:        `t$Path" -Indent $StackLevel -OnlyDebug
    Output-Write " - ArgList:     `t$ArgList" -Indent $StackLevel -OnlyDebug

	$local:process = Start-Process -FilePath $Path -ArgumentList $ArgList -NoNewWindow -PassThru		
	
	if ($Wait)
	{
		$local:process.WaitForExit()
	}
}

##########################################################################

$sidUsers               = "S-1-5-32-545"
$sidDCOMUsers           = "S-1-5-32-562"
$sidAdministrators      = "S-1-5-32-544"

$script:OpcEnumAppID    = "13486D44-4821-11D2-A494-3CB306C10000"
$script:ServerAppID     = "188D08EC-72CB-425B-8EA8-EE2308736440"

$script:SoftwarePath    = $null
$script:SoftwareVersion = $null

$script:ScadaPath       = $null
$script:ServerPath      = $null
$script:OpcEnumPath     = $null


##########################################################################
## AnalyzeParameters
function SearchInstalledSoftware
{
    $StackLevel = 1

    # Check the operating system version. minimum required version - Windows 8
    $local:Major = [System.Environment]::OSVersion.Version.Major
    $local:Minor = [System.Environment]::OSVersion.Version.Minor    
    if ( -not (($local:Major -gt 6) -or ( ($local:Major -ge 6) -and ($local:Minor -ge 2) ))  )
    {
        Output-Warning "Для работы мастера необходима `"Windows 8`" или более новая"
        return $false
    }

    # Search for installed software
    Output-Write "● Поиск установленного программного обеспечения..."
    $ProgressBar.Visible = $true
    $ProgressBar.Value = 0
    try
    {
        $local:RegistryPath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Tekon\SCADA"

        Get-Item -Path $local:RegistryPath  -ErrorAction Stop | Out-Null
        Get-ItemProperty -Path $local:RegistryPath | Select-Object -ExpandProperty "Path" -ErrorAction Stop | Out-Null
        Get-ItemProperty -Path $local:RegistryPath | Select-Object -ExpandProperty "Version" -ErrorAction Stop | Out-Null

        $script:SoftwarePath    = (Get-ItemProperty -Path $local:RegistryPath -Name "Path").Path
        $script:SoftwareVersion = (Get-ItemProperty -Path $local:RegistryPath -Name "Version").Version
        
        $ProgressBar.Value = 25

        Output-Write "- Обнаружено установленное программное обеспечение v$script:SoftwareVersion" -Indent $StackLevel
        Output-Write "  - Path:         `t$script:SoftwarePath" -Indent $StackLevel
        if ($(Test-Path -Path "$script:SoftwarePath\Scada\scada.exe"))
        {
            $script:ScadaPath = "$script:SoftwarePath\Scada\scada.exe"
            Output-Write "  - Scada:      `t$script:ScadaPath" -Indent $StackLevel
        }
        else
        {
            Output-Warning "Файл `"$script:SoftwarePath\Scada\scada.exe`" не найден" -Indent $StackLevel
        }

        $ProgressBar.Value = 50

        if ($(Test-Path -Path "$script:SoftwarePath\OPC Server\opcsrv.exe"))
        {
            $script:ServerPath = "$script:SoftwarePath\OPC Server\opcsrv.exe"
            Output-Write "  - Server:      `t$script:ServerPath" -Indent $StackLevel

            $local:ServicePath = DCOM_GetAppServicePath $script:ServerAppID $StackLevel
            if ($local:ServicePath -ne $null)
            {
                if ($script:ServerPath -ne $local:ServicePath)
                {
                    Output-Warning "Место регистрации службы `"Tekon OPC Server`" не совпадает с местом установки программного обеспечения." -Indent $StackLevel
                                                     
                    Output-Write "  - Service:    `t$local:ServicePath" -Indent $StackLevel

                    $script:ServerPath = $local:ServicePath
                }
            }
            else
            {
                Output-Warning "Cлужба `"Tekon OPC Server`" не зарегистрирована в системе." -Indent $StackLevel
            }

            $ProgressBar.Value = 75

            $local:OpcEnumPath = DCOM_GetAppServicePath $script:OpcEnumAppID $StackLevel
            if ($local:OpcEnumPath -ne $null)
            {
                $script:OpcEnumPath = $local:OpcEnumPath
                Output-Write "  - OpcEnum: `t$script:OpcEnumPath" -Indent $StackLevel
            }
            else
            {
                Output-Warning "Cлужба `"OpcEnum`" не зарегистрирована в системе." -Indent $StackLevel
            }
        }
        else
        {
            Output-Warning "Файл `"$script:SoftwarePath\OPC Server\opcsrv.exe`" не найден" -Indent $StackLevel
        }

        $ProgressBar.Value = 100

        $local:ver = $script:SoftwareVersion.Split(".")
        if ($local:ver.count -eq 3)
        {
            if ($([convert]::ToInt32($local:ver[0])*100*100+[convert]::ToInt32($local:ver[1])*100+[convert]::ToInt32($local:ver[2])) -le 20702)
            {
                Output-Warning "При настойке АСУД Scada v2.7.xx или более старых, на компьютере с OPC сервером должен присутствовать" -Indent $StackLevel
                Output-Write "    пользователь, учетные данные которого совпадают с учетными данными пользователя запускающего АСУД Scada."  -Indent $StackLevel
            }
        }

        return $true
    }
    catch    
    {
        $ProgressBar.Value = 25

        Output-Warning "Программное обеспечение установлено неправильно. Корректная работа не гарантируется!" -Indent $StackLevel

        if ($(Test-Path -Path "c:\1Tekon\ASUD Scada\Scada\scada.exe"))
        {
            $script:ScadaPath = "c:\1Tekon\ASUD Scada\Scada\scada.exe"
            Output-Warning "Обнаружена установленная АСУД Scada." -Indent $StackLevel
            Output-Write "  - Scada:      `t$script:ScadaPath" -Indent $StackLevel
        }

        $ProgressBar.Value = 50

        if ($(Test-Path -Path "c:\1Tekon\ASUD Scada\OPC Server\opcsrv.exe"))
        {
            $script:ServerPath = "c:\1Tekon\ASUD Scada\OPC Server\opcsrv.exe"
            Output-Warning "Обнаружен установленный OPC сервер." -Indent $StackLevel
            Output-Write "  - Server:      `t$script:ServerPath" -Indent $StackLevel

            $local:ServicePath = DCOM_GetAppServicePath $script:ServerAppID $StackLevel
            if ($local:ServicePath -ne $null)
            {
                if ($script:ServerPath -ne $local:ServicePath)
                {
                    Output-Warning "Место регистрации службы `"Tekon OPC Server`" не совпадает с местом обнаружени исполняемого файла `"opcsrv.exe`"." -Indent $StackLevel
                                                     
                    Output-Write "  - Service:    `t$local:ServicePath" -Indent $StackLevel

                    $script:ServerPath = $local:ServicePath
                }
            }
            else
            {
                Output-Warning "Cлужба `"Tekon OPC Server`" не зарегистрирована в системе." -Indent $StackLevel
            }

            $ProgressBar.Value = 75

            $local:OpcEnumPath = DCOM_GetAppServicePath $script:OpcEnumAppID $StackLevel
            if ($local:OpcEnumPath -ne $null)
            {
                $script:OpcEnumPath = $local:OpcEnumPath
                Output-Write "  - OpcEnum: `t$script:OpcEnumPath" -Indent $StackLevel
            }
            else
            {
                Output-Warning "Cлужба `"OpcEnum`" не зарегистрирована в системе." -Indent $StackLevel
            }

            $ProgressBar.Value = 100
        }
        else
        {
            $local:ServicePath = DCOM_GetAppServicePath $script:ServerAppID $StackLevel
            if ($local:ServicePath -ne $null)
            {
                Output-Warning "Обнаружена зарегистрированная в системе служба `"Tekon OPC Server`"." -Indent $StackLevel
                Output-Write "  - Service:    `t$local:ServicePath" -Indent $StackLevel

                $script:ServerPath = $local:ServicePath
            }
            else
            {
                Output-Warning "Cлужба `"Tekon OPC Server`" не зарегистрирована в системе." -Indent $StackLevel
            }
        }

        return $true
    }
}

##########################################################################
## AnalyzeParameters
function CreateDefaultSettings
{
    $script:SelectedWnd                       = $null    
    $script:Settings.ActionsNumber            = 0

    $script:Settings = @{}
    $script:Settings.ScadaPath                = $null
    $script:Settings.ServerPath               = $null
    $script:Settings.OpcEnumPath              = $null

    $script:Settings.ServerUser = @{}
    $script:Settings.ServerUser.Create        = $false
    $script:Settings.ServerUser.Name          = $null
    $script:Settings.ServerUser.Password      = $null
    $script:Settings.ServerUser.AddToGroup    = $null
    $script:Settings.ServerUser.LocalDeny     = $false
    $script:Settings.ServerUser.NetworkAllow  = $false
    $script:Settings.ServerUser.NetworkDeny   = $false
    $script:Settings.ServerUser.LoginAsSvc    = $false

    $script:Settings.ScadaUser = @{}
    $script:Settings.ScadaUser.Create         = $false
    $script:Settings.ScadaUser.Name           = $null
    $script:Settings.ScadaUser.Password       = $null
    $script:Settings.ScadaUser.AddToGroup     = $null
    $script:Settings.ScadaUser.LocalDeny      = $false
    $script:Settings.ScadaUser.NetworkAllow   = $false
    $script:Settings.ScadaUser.ServerControl  = $false
    $script:Settings.ScadaUser.ServerFolder   = $false
    $script:Settings.ScadaUser.OpcEnumControl = $false
    $script:Settings.ScadaUser.AllowDCOM      = $false
    $script:Settings.ScadaUser.AllowServer    = $false
    $script:Settings.ScadaUser.AllowOpcEnum   = $false

    $script:Settings.DCOMUsers = @{}
    $script:Settings.DCOMUsers.AllowDCOM      = $false
    $script:Settings.DCOMUsers.AllowServer    = $false
    $script:Settings.DCOMUsers.AllowOpcEnum   = $false
    $script:Settings.DCOMUsers.ServerControl  = $false
    $script:Settings.DCOMUsers.OpcEnumControl = $false
}

##########################################################################
## AnalyzeParameters
function AnalyzeParameters
{
    $StackLevel = 2

    # resetting the settings to the default values
    CreateDefaultSettings

    # Which software is selected
    if ( -not ($cb_server.Checked -or $cb_scada.Checked) )
    {
        Output-Warning "Необходимо выбрать настраиваемое ПО." -Indent $StackLevel        

        $null = MessageBox "Необходимо выбрать настраиваемое ПО." $main_form.Text $ID_OK $ICON_ERROR

        return $false
    }

    # Clear output if not in debug mode (-not IgnoreOnlyDebug)
    if ( -not $cb_debug.Checked ) { Output-Clear } else {Output-Write " "}
    Output-Write "● Анализ введенных параметров"

    $ProgressBar.Visible = $true
    $ProgressBar.Value = 0

    # Checking the correct path to the 'Scada' executable file
    if ( $cb_scada.Checked )
    {
        if ( $tb_scada_exe.Text.IndexOf(".exe") -lt 0 )
        {
            $tb_scada_exe.Text = $tb_scada_exe.Text.TrimEnd("\")

            if ($(Test-Path -Path "$($tb_scada_exe.Text)\scada.exe"))
            {                
                $tb_scada_exe.Text = "$($tb_scada_exe.Text)\scada.exe"
            }
        }

        if ( -not $(Test-Path -Path $tb_scada_exe.Text) )
        {
            Output-Warning "Файл `"$($tb_scada_exe.Text)`" не найден" -Indent $StackLevel

            $null = MessageBox "Файл `"$($tb_scada_exe.Text)`" не найден" $main_form.Text $ID_OK $ICON_ERROR

            $script:SelectedWnd = $tb_scada_exe

            return $false
        }
        elseif ( $tb_scada_exe.Text.IndexOf(".exe") -lt 0 )
        {
            Output-Warning "Файл `"$($tb_scada_exe.Text)`" не исполняемый" -Indent $StackLevel -OnlyDebug

            $null = MessageBox "Файл `"$($tb_scada_exe.Text)`" не исполняемый" $main_form.Text $ID_OK $ICON_ERROR

            $script:SelectedWnd = $tb_scada_exe

            return $false
        }

        $ProgressBar.Value = 5

        $script:Settings.ScadaPath = $tb_scada_exe.Text
    }


    # Checking the correct path to the 'Tekon OPC Server' executable file
    if ( $cb_server.Checked )
    {
        if ( $tb_server_exe.Text.IndexOf(".exe") -lt 0 )
        {
            $tb_server_exe.Text = $tb_server_exe.Text.TrimEnd("\")

            if ($(Test-Path -Path "$($tb_server_exe.Text)\opcsrv.exe"))
            {                
                $tb_server_exe.Text = "$($tb_server_exe.Text)\opcsrv.exe"
            }
        }

        if ( -not $(Test-Path -Path $tb_server_exe.Text) )
        {
            Output-Warning "Файл `"$($tb_server_exe.Text)`" не найден" -Indent $StackLevel

            $null = MessageBox "Файл `"$($tb_server_exe.Text)`" не найден" $main_form.Text $ID_OK $ICON_ERROR

            $script:SelectedWnd = $tb_server_exe

            return $false
        }
        elseif ( $tb_server_exe.Text.IndexOf(".exe") -lt 0 )
        {
            Output-Warning "Файл `"$($tb_server_exe.Text)`" не исполняемый" -Indent $StackLevel -OnlyDebug

            $null = MessageBox "Файл `"$($tb_server_exe.Text)`" не исполняемый" $main_form.Text $ID_OK $ICON_ERROR
            
            $script:SelectedWnd = $tb_server_exe

            return $false
        }

        $ProgressBar.Value = 10

        $script:Settings.ServerPath = $tb_server_exe.Text
    }


    # Check if OPC Enum is installed
    if ( $cb_server.Checked -or $cb_scada.Checked )
    {
        
        $local:OpcEnumPath = DCOM_GetAppServicePath $script:OpcEnumAppID $StackLevel
        if ( $local:OpcEnumPath -ne $null )
        {
            $ProgressBar.Value = 15
            $script:Settings.OpcEnumPath = $local:OpcEnumPath
        }
        else
        {
            Output-Warning "Cлужба `"OpcEnum`" не зарегистрирована в системе.'" -Indent $StackLevel
            Output-Write   "    Необходимо установить компонент 'OPC Core Components Redistributable (x86) 3.x.x'" -Indent $StackLevel
        }
    }

    # Checking the correctness of entering credentials of the user "Server"
    if ( $cb_server.Checked -or $cb_scada.Checked )
    {
        Output-Write "• Пользователь Server:" -Indent 1 -LinesBefore 1

        if ( -not $(UserMgt_IsUserameValid $tb_server_usr.Text $($StackLevel+1)) )
        {
            Output-Warning "Невозможно использовать имя `"$($tb_server_usr.Text)`". Имя пользователя не может быть пустым," -Indent $StackLevel
            Output-Write "    состоять только из точек или пробелов, а также содержать следующих знаков: /\`"*[]:|<>+=;,?@" -Indent $StackLevel

            $null = MessageBox "Невозможно использовать имя `"$($tb_server_usr.Text)`". Имя пользователя не может быть пустым, состоять только из точек или пробелов, а также содержать следующих знаков:`n/\`"*[]:|<>+=;,?@" "Локальные пользователи и группы" $ID_OK $ICON_ERROR

            $script:SelectedWnd = $tb_server_usr

            return $false
        }

        if ($tb_server_passwd.Text.Length -eq 0)
        {
            $local:SystemInfo = Get-WmiObject -Class Win32_ComputerSystem
            $local:ComputerName = $local:SystemInfo.DNSHostName 

            Output-Warning "Пароль не может быть пустым." -Indent $StackLevel

            $null = MessageBox "Произошла следующая ошибка при попытке создания пользователя `"$($tb_server_usr.Text)`" на компьютере `"$local:ComputerName`":`n`n`nПароль не может быть пустым." "Локальные пользователи и группы" $ID_OK $ICON_ERROR

            $script:SelectedWnd = $tb_server_passwd

            return $false
        }
      
        $ProgressBar.Value = 20

        $script:Settings.ServerUser.Name     = $tb_server_usr.Text
        $script:Settings.ServerUser.Password = ConvertTo-SecureString -String $tb_server_passwd.Text -AsPlainText -Force

        if (UserMgt_IsObjectExist $script:Settings.ServerUser.Name $($StackLevel+1))
        {
            if (-not $(UserMgt_IsCredintalsValid $script:Settings.ServerUser.Name $script:Settings.ServerUser.Password $($StackLevel+1)) )
            {
                $ProgressBar.Value = 25

                $local:SystemInfo = Get-WmiObject -Class Win32_ComputerSystem
                $local:ComputerName = $local:SystemInfo.DNSHostName

                $local:NoError = $false

                $null = LSP_Export $($StackLevel+1)

                    if ( $(LSP_IsValueExist "Privilege Rights" "SeDenyInteractiveLogonRight" $script:Settings.ServerUser.Name $($StackLevel+1)) )
                    {
                        Output-Warning "Невозможно проверить корректность учетных данных пользователя `"$($script:Settings.ServerUser.Name)`"." -Indent $StackLevel
                        Output-Write   "    Пользователю `"$($script:Settings.ServerUser.Name)`" запрещен `"Локальный вход в систему`"." -Indent $StackLevel

                        $local:NoError = $true
                    }
                    else
                    {
                        try
                        {
                            $local:up = [System.DirectoryServices.AccountManagement.UserPrincipal]::FindByIdentity(
                                [System.DirectoryServices.AccountManagement.UserPrincipal]::Current.Context, $script:Settings.ServerUser.Name)

                            if ($local:up -ne $null)
                            {
                                for ($i=0; $i -lt $up.GetGroups().SID.Value.count; $i++)
                                {
                                    $local:GroupSID = if ($up.GetGroups().SID.Value.count -eq 1) {$up.GetGroups().SID.Value} else {$up.GetGroups().SID.Value[$i]}

                                    if ( $(LSP_IsValueExist "Privilege Rights" "SeDenyInteractiveLogonRight" "*$local:GroupSID" $($StackLevel+1)) )
                                    {
                                        $local:GroupName = $(UserMgt_GetName $local:GroupSID $($StackLevel+1))

                                        if ( $local:GroupName -ne $null)
                                        {
                                            Output-Warning "Невозможно проверить корректность учетных данных пользователя `"$($script:Settings.ServerUser.Name)`"." -Indent $StackLevel
                                            Output-Write ` "    Пользователь `"$($script:Settings.ServerUser.Name)`" состоит в группе `"$local:GroupName`" которой запрещен `"Локальный вход в систему`"." -Indent $StackLevel

                                            $local:NoError = $true
                                        }
                                    }
                                }
                            }
                        }
                        catch
                        {}
                    }


                    if ( $(LSP_IsValueExist "Privilege Rights" "SeDenyNetworkLogonRight" $script:Settings.ServerUser.Name $($StackLevel+1)) )
                    {
                        Output-Warning "Невозможно проверить корректность учетных данных пользователя `"$($script:Settings.ServerUser.Name)`"." -Indent $StackLevel
                        Output-Write   "    Пользователю `"$($script:Settings.ServerUser.Name)`" отказано в`"Доступе к компьютеру из сети`"." -Indent $StackLevel

                        $local:NoError = $true
                    }
                    else
                    {
                        try
                        {
                            $local:up = [System.DirectoryServices.AccountManagement.UserPrincipal]::FindByIdentity(
                                [System.DirectoryServices.AccountManagement.UserPrincipal]::Current.Context, $script:Settings.ServerUser.Name)

                            if ($local:up -ne $null)
                            {
                                for ($i=0; $i -lt $up.GetGroups().SID.Value.count; $i++)
                                {
                                    $local:GroupSID = if ($up.GetGroups().SID.Value.count -eq 1) {$up.GetGroups().SID.Value} else {$up.GetGroups().SID.Value[$i]}

                                    if ( $(LSP_IsValueExist "Privilege Rights" "SeDenyNetworkLogonRight" "*$local:GroupSID" $($StackLevel+1)) )
                                    {
                                        $local:GroupName = $(UserMgt_GetName $local:GroupSID $($StackLevel+1))

                                        if ( $local:GroupName -ne $null)
                                        {
                                            Output-Warning "Невозможно проверить корректность учетных данных пользователя `"$($script:Settings.ServerUser.Name)`"." -Indent $StackLevel
                                            Output-Write ` "    Пользователь `"$($script:Settings.ServerUser.Name)`" состоит в группе `"$local:GroupName`" которой отказано в `"Доступе к компьютеру из сети`"." -Indent $StackLevel

                                            $local:NoError = $true
                                        }
                                    }
                                }
                            }
                        }
                        catch
                        {}
                    }


                LSP_Rollback $($StackLevel+1)


                if (-not $local:NoError)
                {                    
                    Output-Warning "Учетная запись `"$($script:Settings.ServerUser.Name)`" уже существует. Пароль введен неправильно." -Indent $StackLevel

                    $null = MessageBox "Произошла следующая ошибка при попытке создания пользователя `"$($script:Settings.ServerUser.Name)`" на компьютере `"$local:ComputerName`"`n`nУчетная запись уже существует.`nПароль введен неправильно." "Локальные пользователи и группы" $ID_OK $ICON_ERROR
                    
                    $script:SelectedWnd = $tb_server_passwd

                    return $false
                }
            }
             
            $ProgressBar.Value = 30

            Output-Write " - Учетная запись `"$($script:Settings.ServerUser.Name)`" уже существует" -Indent $StackLevel

            $local:GroupSID = if ($cb_server.Checked) {$sidAdministrators} else {$sidUsers}
            if ( -not $(UserMgt_IsMemberOfGroup $script:Settings.ServerUser.Name $(UserMgt_GetName $local:GroupSID $($StackLevel+1)) $($StackLevel+1)) )
            {                
                $script:Settings.ActionsNumber += 1
                $script:Settings.ServerUser.AddToGroup = $local:GroupSID

                $local:GroupName = $(UserMgt_GetName $script:Settings.ServerUser.AddToGroup  $($StackLevel+1))

                if ($local:GroupName -ne $null)
                {
                    Output-Write " ▸ Пользователь  `"$($script:Settings.ServerUser.Name)`" будет добавлен в группу '$local:GroupName'" -Indent $StackLevel                    
                }
            }            
        }
        else
        {
            $ProgressBar.Value = 30

            $script:Settings.ActionsNumber += 1
            $script:Settings.ServerUser.Create = $true            
            Output-Write "▸ Будет создан новый пользователь `"$($script:Settings.ServerUser.Name)`"" -Indent $StackLevel

            
            $local:GroupSID = if ($cb_server.Checked) {$sidAdministrators} else {$sidUsers}

            $script:Settings.ActionsNumber += 1
            $script:Settings.ServerUser.AddToGroup = $local:GroupSID
            $local:GroupName = $(UserMgt_GetName $script:Settings.ServerUser.AddToGroup $($StackLevel+1))
            if ($local:GroupName -ne $null)
            {
                Output-Write " ▸ Пользователь  `"$($script:Settings.ServerUser.Name)`" будет добавлен в группу '$local:GroupName'" -Indent $StackLevel
            }           
        }

        # ServerUser Local deny
        if ( $cb_server_localdeny.Checked )
        {
            $local:LoggedInUserName = UserMgt_GetLoggedIn $($StackLevel+1)
            if ($local:LoggedInUserName -ne $null)
            {
                if ($script:Settings.ServerUser.Name.Split("\")[-1] -eq $(UserMgt_GetLoggedIn $($StackLevel+1)))
                {
                    Output-Warning "Нельзя запретить `"Локальный вход в систему`" пользователю, вошедшему в систему." -Indent $StackLevel
                    Output-Write ` "    Это может привести к невозможности использования компьютера после перезагрузки." -Indent $StackLevel

                    $cb_server_localdeny.Checked = $false
                }
                else
                {
                    $script:Settings.ActionsNumber += 1
                    $script:Settings.ServerUser.LocalDeny = $true            

                    Output-Write " ▸ Пользователю `"$($script:Settings.ServerUser.Name)`" будет запрещен `"Локальный вход в систему`"" -Indent $StackLevel
                }
            }
            else
            {
                Output-Warning "Не удалось получить имя пользователя вошедшего в систему." -Indent $StackLevel
                Output-Write "    Не рекомендуется использование данного мастера через Удаленный рабочий стол." -Indent $StackLevel
            }                        
        }
        else { $script:Settings.ServerUser.LocalDeny = $false }

        $ProgressBar.Value = 35

        # ServerUser Network allow/deny
        if ( $cb_scada.Checked )
        {
            # if scada is selected - can not be denied, can only be allowed
            $script:Settings.ActionsNumber += 1
            $script:Settings.ServerUser.NetworkDeny  = $false
            $script:Settings.ServerUser.NetworkAllow = $true

            Output-Write " ▸ Пользователю `"$($script:Settings.ServerUser.Name)`" будет разрешен `"Доступ к компьютеру из сети`"" -Indent $StackLevel
        }
        else
        {
            # ServerUser Network deny
            if ( $cb_server_networkdeny.Checked )
            {
                $script:Settings.ActionsNumber += 1
                $script:Settings.ServerUser.NetworkDeny = $true

                Output-Write " ▸ Пользователю `"$($script:Settings.ServerUser.Name)`" будет отказано в `"Доступе к этому компьютеру из сети`"" -Indent $StackLevel
            }
            else { $script:Settings.ServerUser.NetworkDeny = $false }
        }

        $ProgressBar.Value = 40

        # ServerUser LoginAsSvc
        if ($cb_server.Checked)
        {
            $script:Settings.ActionsNumber += 1
            $script:Settings.ServerUser.LoginAsSvc = $true
           
            Output-Write " ▸ Пользователю `"$($script:Settings.ServerUser.Name)`" будет разрешен `"Вход в качестве службы`"" -Indent $StackLevel           
        }

        $ProgressBar.Value = 45
    }


    # Checking the correctness of entering credentials of the user "Scada"
    if ( $cb_server.Checked )
    {
        Output-Write "• Пользователь Scada:" -Indent 1 -LinesBefore 1

        if ( -not $(UserMgt_IsUserameValid $tb_scada_usr.Text $($StackLevel+1)) )
        {
            Output-Warning "Невозможно использовать имя `"$($tb_scada_usr.Text)`". Имя пользователя не может быть пустым, состоять только из точек или пробелов, а также содержать следующих знаков: /\`"*[]:|<>+=;,?@" -Indent $StackLevel

            $null = MessageBox "Невозможно использовать имя `"$($tb_scada_usr.Text)`". Имя пользователя не может быть пустым, состоять только из точек или пробелов, а также содержать следующих знаков:`n/\`"*[]:|<>+=;,?@" "Локальные пользователи и группы" $ID_OK $ICON_ERROR

            $script:SelectedWnd = $tb_scada_usr

            return $false
        }

        if ($tb_scada_passwd.Text.Length -eq 0)
        {
            $local:SystemInfo = Get-WmiObject -Class Win32_ComputerSystem
            $local:ComputerName = $local:SystemInfo.DNSHostName 

            Output-Warning "Пароль не может быть пустым." -Indent $StackLevel

            $null = MessageBox "Произошла следующая ошибка при попытке создания пользователя `"$($tb_scada_usr.Text)`" на компьютере `"$local:ComputerName`":`n`n`nПароль не может быть пустым." "Локальные пользователи и группы" $ID_OK $ICON_ERROR

            $script:SelectedWnd = $tb_scada_passwd

            return $false
        }
      
        $ProgressBar.Value = 50

        $script:Settings.ScadaUser.Name     = $tb_scada_usr.Text
        $script:Settings.ScadaUser.Password = ConvertTo-SecureString -String $tb_scada_passwd.Text -AsPlainText -Force

        if (UserMgt_IsObjectExist $script:Settings.ScadaUser.Name $($StackLevel+1))
        {
            if (-not $(UserMgt_IsCredintalsValid $script:Settings.ScadaUser.Name $script:Settings.ScadaUser.Password $($StackLevel+1)) )
            {
                $ProgressBar.Value = 55

                $local:SystemInfo = Get-WmiObject -Class Win32_ComputerSystem
                $local:ComputerName = $local:SystemInfo.DNSHostName

                $local:NoError = $false

                $null = LSP_Export $($StackLevel+1)

                    if ( $(LSP_IsValueExist "Privilege Rights" "SeDenyInteractiveLogonRight" $script:Settings.ScadaUser.Name $($StackLevel+1)) )
                    {
                        Output-Warning "Невозможно проверить корректность учетных данных пользователя `"$($script:Settings.ScadaUser.Name)`"." -Indent $StackLevel                    
                        Output-Write   "    Пользователю `"$($script:Settings.ScadaUser.Name)`" запрещен `"Локальный вход в систему`"." -Indent $StackLevel

                        $local:NoError = $true
                    }
                    else
                    {
                        try
                        {
                            $local:up = [System.DirectoryServices.AccountManagement.UserPrincipal]::FindByIdentity(
                                [System.DirectoryServices.AccountManagement.UserPrincipal]::Current.Context, $script:Settings.ScadaUser.Name)

                            if ($local:up -ne $null)
                            {
                                for ($i=0; $i -lt $up.GetGroups().SID.Value.count; $i++)
                                {
                                    $local:GroupSID = if ($up.GetGroups().SID.Value.count -eq 1) {$up.GetGroups().SID.Value} else {$up.GetGroups().SID.Value[$i]}

                                    if ( $(LSP_IsValueExist "Privilege Rights" "SeDenyInteractiveLogonRight" "*$local:GroupSID" $($StackLevel+1)) )
                                    {
                                        $local:GroupName = $(UserMgt_GetName $local:GroupSID $($StackLevel+1))

                                        if ( $local:GroupName -ne $null)
                                        {
                                            Output-Warning "Невозможно проверить корректность учетных данных пользователя `"$($script:Settings.ScadaUser.Name)`"." -Indent $StackLevel
                                            Output-Write ` "    Пользователь `"$($script:Settings.ScadaUser.Name)`" состоит в группе `"$local:GroupName`" которой запрещен `"Локальный вход в систему`"." -Indent $StackLevel

                                            $local:NoError = $true
                                        }
                                    }
                                }
                            }
                        }
                        catch
                        {}
                    }

                    if ( $(LSP_IsValueExist "Privilege Rights" "SeDenyNetworkLogonRight" $script:Settings.ScadaUser.Name $($StackLevel+1)) )
                    {
                        Output-Warning "Невозможно проверить корректность учетных данных пользователя `"$($script:Settings.ScadaUser.Name)`"." -Indent $StackLevel
                        Output-Write   "    Пользователю `"$($script:Settings.ScadaUser.Name)`" отказано `"Доступе к компьютеру из сети`"." -Indent $StackLevel

                        $local:NoError = $true
                    }
                    else
                    {
                        try
                        {
                            $local:up = [System.DirectoryServices.AccountManagement.UserPrincipal]::FindByIdentity(
                                [System.DirectoryServices.AccountManagement.UserPrincipal]::Current.Context, $script:Settings.ScadaUser.Name)

                            if ($local:up -ne $null)
                            {
                                for ($i=0; $i -lt $up.GetGroups().SID.Value.count; $i++)
                                {
                                    $local:GroupSID = if ($up.GetGroups().SID.Value.count -eq 1) {$up.GetGroups().SID.Value} else {$up.GetGroups().SID.Value[$i]}

                                    if ( $(LSP_IsValueExist "Privilege Rights" "SeDenyNetworkLogonRight" "*$local:GroupSID" $($StackLevel+1)) )
                                    {
                                        $local:GroupName = $(UserMgt_GetName $local:GroupSID $($StackLevel+1))

                                        if ( $local:GroupName -ne $null)
                                        {
                                            Output-Warning "Невозможно проверить корректность учетных данных пользователя `"$($script:Settings.ScadaUser.Name)`"." -Indent $StackLevel
                                            Output-Write ` "    Пользователь `"$($script:Settings.ScadaUser.Name)`" состоит в группе `"$local:GroupName`" которой отказано `"Доступе к компьютеру из сети`"." -Indent $StackLevel

                                            $local:NoError = $true
                                        }
                                    }
                                }
                            }
                        }
                        catch
                        {}
                    }

                LSP_Rollback $($StackLevel+1)
                

                if (-not $local:NoError)
                {                    
                    Output-Warning "Учетная запись `"$($script:Settings.ScadaUser.Name)`" уже существует. Пароль введен неправильно." -Indent $StackLevel

                    $null = MessageBox "Произошла следующая ошибка при попытке создания пользователя `"$($script:Settings.ScadaUser.Name)`" на компьютере `"$local:ComputerName`"`n`nУчетная запись уже существует.`nПароль введен неправильно." "Локальные пользователи и группы" $ID_OK $ICON_ERROR
                    
                    $script:SelectedWnd = $tb_server_passwd

                    return $false
                }
            }
             
            $ProgressBar.Value = 60

            Output-Write " - Учетная запись `"$($script:Settings.ScadaUser.Name)`" уже существует" -Indent $StackLevel

            $local:GroupSID = $sidDCOMUsers
            if ( -not $(UserMgt_IsMemberOfGroup $script:Settings.ScadaUser.Name $(UserMgt_GetName $local:GroupSID $($StackLevel+1)) $($StackLevel+1)) )
            {                
                $script:Settings.ActionsNumber += 1
                $script:Settings.ScadaUser.AddToGroup = $local:GroupSID

                $local:GroupName = $(UserMgt_GetName $script:Settings.ScadaUser.AddToGroup  $($StackLevel+1))

                if ($local:GroupName -ne $null)
                {
                    Output-Write " ▸ Пользователь `"$($script:Settings.ScadaUser.Name)`" будет добавлен в группу '$local:GroupName'" -Indent $StackLevel                    
                }
            }
        }
        else
        {
            $ProgressBar.Value = 60

            $script:Settings.ActionsNumber += 1
            $script:Settings.ScadaUser.Create = $true
            Output-Write "▸ Будет создан новый пользователь `"$($script:Settings.ScadaUser.Name)`"" -Indent $StackLevel


            $local:GroupSID = $sidDCOMUsers

            $script:Settings.ActionsNumber += 1
            $script:Settings.ScadaUser.AddToGroup = $local:GroupSID
            $local:GroupName = $(UserMgt_GetName $script:Settings.ScadaUser.AddToGroup $($StackLevel+1))
            if ($local:GroupName -ne $null)
            {
                Output-Write " ▸ Пользователь  `"$($script:Settings.ScadaUser.Name)`" будет добавлен в группу '$local:GroupName'" -Indent $StackLevel
            }
        }


        $ProgressBar.Value = 65

        # ScadaUser Local deny
        if ( $cb_scada_localdeny.Checked )
        {
            $local:LoggedInUserName = UserMgt_GetLoggedIn $($StackLevel+1)
            if ($local:LoggedInUserName -ne $null)
            {
                if ($script:Settings.ScadaUser.Name.Split("\")[-1] -eq $(UserMgt_GetLoggedIn $($StackLevel+1)))
                {
                    Output-Warning "Нельзя запретить `"Локальный вход в систему`" пользователю, вошедшему в систему." -Indent $StackLevel
                    Output-Write ` "    Это может привести к невозможности использования компьютера после перезагрузки." -Indent $StackLevel

                    $cb_scada_localdeny.Checked = $false
                }
                else
                {
                    $script:Settings.ActionsNumber += 1
                    $script:Settings.ScadaUser.LocalDeny = $true            

                    Output-Write " ▸ Пользователю `"$($script:Settings.ScadaUser.Name)`" будет запрещен `"Локальный вход в систему`"" -Indent $StackLevel
                }  
            }
            else
            {
                Output-Warning "Не удалось получить имя пользователя вошедшего в систему." -Indent 1
                Output-Write "    Не рекомендуется использование данного мастера через Удаленный рабочий стол." -Indent 1
            }
        }
        else { $script:Settings.ScadaUser.LocalDeny = $false }

        $ProgressBar.Value = 70

        # ScadaUser Network allow/deny
        if ( $cb_server.Checked )
        {
            # if scada is selected - can not be denied, can only be allowed
            $script:Settings.ActionsNumber += 1
            $script:Settings.ScadaUser.NetworkAllow = $true

            Output-Write " ▸ Пользователю `"$($script:Settings.ScadaUser.Name)`" будет разрешен `"Доступ компьютеру из сети`"" -Indent $StackLevel
        }
        
        $ProgressBar.Value = 75

        # Grant ScadaUser the rights to control Tekon OPC Server
        if ( $script:Settings.ServerPath -ne $null )
        {
            $script:Settings.ActionsNumber += 1
            $script:Settings.ScadaUser.ServerControl = $true

            $local:AppName = DCOM_GetAppServiceName $script:ServerAppID $($StackLevel+1)
            if ( $local:AppName -ne $null )
            {
                Output-Write " ▸ Пользователю `"$($script:Settings.ScadaUser.Name)`" будут предоставлены права на управление службой `"$local:AppName`"" -Indent $StackLevel
            }
        }

        # Grant ScadaUser access rights to the folder
        if ( $script:Settings.ServerPath -ne $null )
        {
            $local:Folder = split-path -Parent $script:Settings.ServerPath

            if ( $local:Folder -ne $null )
            {
                $script:Settings.ActionsNumber += 1
                $script:Settings.ScadaUser.ServerFolder = $true

                Output-Write " ▸ Пользователю `"$($script:Settings.ScadaUser.Name)`" будут предоставлены разрешения на полный доступ к каталогу  `"$local:Folder`"" -Indent $StackLevel
            }
        }

        $ProgressBar.Value = 80

        # Grant ScadaUser the rights to control OPC Enum
        if ( $script:Settings.OpcEnumPath -ne $null )
        {
            $script:Settings.ActionsNumber += 1
            $script:Settings.ScadaUser.OpcEnumControl = $true

            $local:AppName = DCOM_GetAppServiceName $script:OpcEnumAppID $($StackLevel+1)
            if ( $local:AppName -ne $null )
            {
                Output-Write " ▸ Пользователю `"$($script:Settings.ScadaUser.Name)`" будут предоставлены права на управление службой `"$local:AppName`"" -Indent $StackLevel
            }            
        }
        

        $ProgressBar.Value = 85

        # Setting up access rights via DСOM                
        $script:Settings.ActionsNumber += 1
        $script:Settings.ScadaUser.AllowDCOM = $true
        Output-Write " ▸ Пользователю `"$($script:Settings.ScadaUser.Name)`" будут предоставлены права на доступ к `"DCOM компонентам`" на этом компьютере" -Indent $StackLevel

        if ($local:ServerAppName -ne $null)
        {
            $local:ServerAppName = $(DCOM_GetAppName $script:ServerAppID $StackLevel)
            $script:Settings.ActionsNumber += 1
            $script:Settings.ScadaUser.AllowServer = $true        
                   
            Output-Write " ▸ Пользователю `"$($script:Settings.ScadaUser.Name)`" будут предоставлены права на доступ к `"$local:ServerAppName`"" -Indent $StackLevel
        }        


        if ($local:OpcEnumAppName -ne $null)
        {
            $local:OpcEnumAppName = $(DCOM_GetAppName $script:OpcEnumAppID $StackLevel)
            $script:Settings.ActionsNumber += 1
            $script:Settings.ScadaUser.AllowOpcEnum = $true    
                       
            Output-Write " ▸ Пользователю `"$($script:Settings.ScadaUser.Name)`" будут предоставлены права на доступ к `"$local:OpcEnumAppName`"" -Indent $StackLevel
        }

        $local:GroupName = $(UserMgt_GetName $sidDCOMUsers $($StackLevel+1))
        if ($local:GroupName -ne $null)
        {
            Output-Write "• Группа $($local:GroupName):" -Indent 1 -LinesBefore 1
            
            # Grant DCOM users the rights to control Tekon OPC Server
            if ( $script:Settings.ServerPath -ne $null )
            {
                $script:Settings.ActionsNumber += 1
                $script:Settings.DCOMUsers.ServerControl = $true

                $local:AppName = DCOM_GetAppServiceName $script:ServerAppID $($StackLevel+1)
                if ( $local:AppName -ne $null )
                {
                    Output-Write " ▸ Группе `"$($local:GroupName)`" будут предоставлены права на управление службой `"$local:AppName`"" -Indent $StackLevel
                }
            }

            $ProgressBar.Value = 87

            # Grant DCOM users the rights to control OPC Enum
            if ( $script:Settings.OpcEnumPath -ne $null )
            {
                $script:Settings.ActionsNumber += 1
                $script:Settings.DCOMUsers.OpcEnumControl = $true

                $local:AppName = DCOM_GetAppServiceName $script:OpcEnumAppID $($StackLevel+1)
                if ( $local:AppName -ne $null )
                {
                    Output-Write " ▸ Группе `"$($local:GroupName)`" будут предоставлены права на управление службой `"$local:AppName`"" -Indent $StackLevel
                }
            }


            $script:Settings.ActionsNumber += 1
            $script:Settings.DCOMUsers.AllowDCOM = $true

            Output-Write " ▸ Группе `"$local:GroupName`" будут предоставлены права на доступ к `"DCOM компонентам`" на этом компьютере" -Indent $StackLevel
            if ($local:ServerAppName -ne $null)
            {
                $script:Settings.ActionsNumber += 1
                $script:Settings.DCOMUsers.AllowServer = $true

                Output-Write " ▸ Группе `"$local:GroupName`" будут предоставлены права на доступ к `"$local:ServerAppName`"" -Indent $StackLevel
            }
         
            if ($local:OpcEnumAppName -ne $null)
            {                
                $script:Settings.ActionsNumber += 1
                $script:Settings.DCOMUsers.AllowOpcEnum = $true

                Output-Write " ▸ Группе `"$local:GroupName`" будут предоставлены права на доступ к `"$local:OpcEnumAppName`"" -Indent $StackLevel
            }
        }
    }

    $ProgressBar.Value = 90


    # Configuring the Firewall
    Output-Write "• Настрока Брандмауэра:" -Indent 1 -LinesBefore 1

    $local:ComputerName = $(UserMgt_GetComputerName $StackLevel)
    if ( $local:ComputerName -ne $null)
    {
        Output-Write "- $($local:ComputerName): " -Indent $StackLevel   
    }

    $script:Settings.ActionsNumber += 1
    Output-Write " ▸ Будет разрешен входящий трафик ICMP4 (Ping)" -Indent $StackLevel

    $script:Settings.ActionsNumber += 1
    Output-Write " ▸ Будет разрешен входящий трафик TCP на порт 135 (DCOM)" -Indent $StackLevel
    
    $ProgressBar.Value = 95

    if ($script:Settings.ScadaPath -ne $null)
    {        
        Output-Write "- АСУД Scada: `"$($script:Settings.ScadaPath)`"" -Indent $StackLevel
        
        $script:Settings.ActionsNumber += 1
        Output-Write " ▸ Для `"$(Split-Path $script:Settings.ScadaPath -Leaf)`" будет создано правило брандмауэра, разрешающие входяшие подключения" -Indent $StackLevel
        
        $script:Settings.ActionsNumber += 1
        Output-Write " ▸ Для `"$(Split-Path $script:Settings.ScadaPath -Leaf)`" будет создано правило брандмауэра, разрешающие исходящие подключения" -Indent $StackLevel
    }

    if ($script:Settings.ServerPath -ne $null)
    {
        Output-Write "- OPC сервер: `"$($script:Settings.ServerPath)`"" -Indent $StackLevel
        
        $script:Settings.ActionsNumber += 1
        Output-Write " ▸ Для `"$(Split-Path $script:Settings.ServerPath -Leaf)`" будет создано правило брандмауэра, разрешающие входяшие подключения" -Indent $StackLevel
        
        $script:Settings.ActionsNumber += 1
        Output-Write " ▸ Для `"$(Split-Path $script:Settings.ServerPath -Leaf)`" будет создано правило брандмауэра, разрешающие исходящие подключения" -Indent $StackLevel
    }

    if ($script:Settings.OpcEnumPath -ne $null)
    {
        Output-Write "- OPC Enum: `"$($script:Settings.OpcEnumPath)`"" -Indent $StackLevel

        $script:Settings.ActionsNumber += 1
        Output-Write " ▸ Для `"$(Split-Path $script:Settings.OpcEnumPath -Leaf)`" будет создано правило брандмауэра, разрешающие входяшие подключения" -Indent $StackLevel
    }



    if ($script:Settings.ServerPath -ne $null)
    {
        # Configuring the Service

        $local:ServiceName = DCOM_GetAppServiceName $script:ServerAppID $StackLevel
        if ($local:ServiceName -ne $null)
        {
            # Configuring the Firewall
            Output-Write "• Настрока службы `"$local:ServiceName`":" -Indent 1 -LinesBefore 1

            $script:Settings.ActionsNumber += 1
            Output-Write " ▸ Будет настроена на вход в систему с учетной записью: `"$($script:Settings.ServerUser.Name)`"" -Indent $StackLevel
        } 
    }

    $ProgressBar.Value = 100

    return $true
}




##########################################################################
## ApplySettings

function ProgressBarInc
{
    if ($script:Settings.ActionsNumber -gt 0)
    {
        $local:Value = $ProgressBar.Value + (100 / $script:Settings.ActionsNumber)
        if ($local:Value -le 100)
        {
            $ProgressBar.Value = $ProgressBar.Value + (100 / $script:Settings.ActionsNumber)
        }
        else
        {
            $ProgressBar.Value = 100
        }
    }
}

function ApplySettings
{
    $StackLevel = 2
            
    # Clear output if not in debug mode (-not IgnoreOnlyDebug)
    if ( -not $cb_debug.Checked ) { Output-Clear } else {Output-Write " "}

    Output-Write "● Применение параметров" -LinesAfter 1

    Output-Write "- Settings: $($global:Settings | ConvertTo-Json -Compress)" -Indent $StackLevel -OnlyDebug

    $ProgressBar.Visible = $true
    $ProgressBar.Value = 0


    ### Setting up users
    if ( $script:Settings.ServerUser.Create -or ($script:Settings.ServerUser.AddToGroup -ne $null) -or`
         $script:Settings.ScadaUser.Create  -or ($script:Settings.ScadaUser.AddToGroup -ne $null) )
    {
        Output-Write "• Настрока пользователей:" -Indent 1 -LinesBefore 1
    }

    # ServerUser::Create
    if ( $script:Settings.ServerUser.Create )
    {
        $local:userdesc = if ($script:Settings.ServerPath -ne $null) {"Учетная запись пользователя, которая используется для запуска службы Tekon OPC Server."} `
                          else {"Учетная запись пользователя, которая используется для олицетворения обратных вызовов от Tekon OPC Server."}

        $local:result = UserMgt_CreateUser $script:Settings.ServerUser.Name $script:Settings.ServerUser.Password `
                                           $script:Settings.ServerUser.Name  $local:userdesc $($StackLevel+1)

        Output-Write "• Создать пользователя `"$($script:Settings.ServerUser.Name)`"" -Indent $StackLevel
        Output-Result $local:result -Indent $($StackLevel+1)
        ProgressBarInc
    }

    # ServerUser::AddToGroup
    if ( $script:Settings.ServerUser.AddToGroup -ne $null )
    {
        $local:GroupName = $(UserMgt_GetName $script:Settings.ServerUser.AddToGroup $($StackLevel+1))
        if ($local:GroupName -ne $null)
        {
            Output-Write "• Добавить пользователя `"$($script:Settings.ServerUser.Name)`" в группу `"$($local:GroupName)`"" -Indent $StackLevel

            if ( $(UserMgt_IsObjectExist $script:Settings.ServerUser.Name $($StackLevel+1)) )
            {
                $local:result =  UserMgt_AddToGroup $script:Settings.ServerUser.Name $local:GroupName $($StackLevel+1)

                Output-Result $local:result -Indent $($StackLevel+1)
            }
            else
            {
                Output-Warning "Пользователь `"$($script:Settings.ServerUser.Name)`" не существует" -Indent $StackLevel
            }    
        }
        else
        {
            Output-Warning "Не удалось добавить пользователя `"$($script:Settings.ServerUser.Name)`" в группу `"sid:$($script:Settings.ServerUser.AddToGroup)`"" -Indent 1
        }

        ProgressBarInc
    }

    # ScadaUser::Create
    if ( $script:Settings.ScadaUser.Create )
    {
        $local:result = UserMgt_CreateUser $script:Settings.ScadaUser.Name $script:Settings.ScadaUser.Password `
            $script:Settings.ScadaUser.Name "Учетная запись пользователя, которая используется для подключения к Tekon OPC Server." $($StackLevel+1)

        Output-Write "• Создать пользователя `"$($script:Settings.ScadaUser.Name)`"" -Indent $StackLevel
        Output-Result $local:result -Indent $($StackLevel+1)
        ProgressBarInc
    }

    # ScadaUser::AddToGroup
    if ( $script:Settings.ScadaUser.AddToGroup -ne $null )
    {
        $local:GroupName = $(UserMgt_GetName $script:Settings.ScadaUser.AddToGroup $($StackLevel+1))
        if ($local:GroupName -ne $null)
        {
            Output-Write "• Добавить пользователя `"$($script:Settings.ScadaUser.Name)`" в группу `"$($local:GroupName)`"" -Indent $StackLevel

            if ( $(UserMgt_IsObjectExist $script:Settings.ScadaUser.Name $($StackLevel+1)) )
            {
                $local:result =  UserMgt_AddToGroup $script:Settings.ScadaUser.Name $local:GroupName $($StackLevel+1)

                Output-Result $local:result -Indent $($StackLevel+1)
            }
            else
            {
                Output-Warning "Пользователь `"$($script:Settings.ScadaUser.Name)`" не существует" -Indent $StackLevel
            }    
        }
        else
        {
            Output-Warning "Не удалось добавить пользователя `"$($script:Settings.ScadaUser.Name)`" в группу `"sid:$($script:Settings.ScadaUser.AddToGroup)`"" -Indent 1
        }

        ProgressBarInc
    }




    ### Local Security Policy
    Output-Write "• Настрока Локальной политики безопасности:" -Indent 1 -LinesBefore 1

    Output-Write "• Экспортировать настройки `"Локальной политики безопасности`"" -Indent $StackLevel
    $local:result = LSP_Export $($StackLevel+1)
    Output-Result $local:result -Indent $($StackLevel+1)

        # ServerUser::DenyLogonLocally
        if ( $script:Settings.ServerUser.LocalDeny )
        {
            Output-Write "• Запретить пользователю `"$($script:Settings.ServerUser.Name)`" `"Локальный вход в систему`"" -Indent $StackLevel

            $local:result = LSP_DenyLogonLocally -User $script:Settings.ServerUser.Name -Stack $($StackLevel+1)
            Output-Result $local:result -Indent $($StackLevel+1)

            ProgressBarInc 
        }

        # ServerUser::DenyAccessFromNetwork
        if ( $script:Settings.ServerUser.NetworkDeny )
        {
            Output-Write "• Отказать пользователю `"$($script:Settings.ServerUser.Name)`" в `"Доступе к компьютеру из сети`"" -Indent $StackLevel

            $local:result = LSP_DenyAccessFromNetwork -User $script:Settings.ServerUser.Name -Stack $($StackLevel+1)
            Output-Result $local:result -Indent $($StackLevel+1)

            ProgressBarInc 
        }


        # ServerUser::GrantServiceLogonRight
        if ( $script:Settings.ServerUser.LoginAsSvc )
        {
            Output-Write "• Разрешить пользователю `"$($script:Settings.ServerUser.Name)`" `"Вход в качестве службы`"" -Indent $StackLevel

            $local:result = LSP_GrantServiceLogonRight -User $script:Settings.ServerUser.Name -Stack $($StackLevel+1)
            Output-Result $local:result -Indent $($StackLevel+1)

            ProgressBarInc 
        }


        # ScadaUser::DenyLogonLocally
        if ( $script:Settings.ScadaUser.LocalDeny )
        {
            Output-Write "• Запретить пользователю `"$($script:Settings.ScadaUser.Name)`" `"Локальный вход в систему`"" -Indent $StackLevel

            $local:result = LSP_DenyLogonLocally -User $script:Settings.ScadaUser.Name -Stack $($StackLevel+1)
            Output-Result $local:result -Indent $($StackLevel+1)

            ProgressBarInc 
        }

        # ServerUser::AllowAccessFromNetwork
        if ( $script:Settings.ServerUser.NetworkAllow )
        {
            Output-Write "• Разрешить пользователю `"$($script:Settings.ServerUser.Name)`" `"Доступ к компьютеру из сети`"" -Indent $StackLevel

            $local:result = LSP_AllowAccessFromNetwork -User $script:Settings.ServerUser.Name -Stack $($StackLevel+1)
            Output-Result $local:result -Indent $($StackLevel+1)

            ProgressBarInc 
        }

        # ScadaUser::AllowAccessFromNetwork
        if ( $script:Settings.ScadaUser.NetworkAllow )
        {
            Output-Write "• Разрешить пользователю `"$($script:Settings.ScadaUser.Name)`" `"Доступ к компьютеру из сети`"" -Indent $StackLevel

            $local:result = LSP_AllowAccessFromNetwork -User $script:Settings.ScadaUser.Name -Stack $($StackLevel+1)
            Output-Result $local:result -Indent $($StackLevel+1)

            ProgressBarInc 
        }

    Output-Write "• Подтвердить изменения `"Локальной политики безопасности`"" -Indent $StackLevel
    $local:result = LSP_CommitChanges $($StackLevel+1)
    Output-Result $local:result -Indent $($StackLevel+1)



    ### Configuring access rights
    Output-Write "• Настройка прав доступа:" -Indent 1 -LinesBefore 1
    
    # Переригестрировать службу
    if ( $script:Settings.ServerPath -ne $null)
    {
        if ( -not $(DCOM_IsAppExist $script:ServerAppID $($StackLevel+1)) )
        {
            Output-Warning "Служба `"AppID:$($script:ServerAppID)`" не зарегистрирована в системе" -Indent 1
            Output-Write "• Рагистрация службы `"AppID:$($script:ServerAppID)`"" -Indent $StackLevel

            $local:ArgList = "-s -r --wait --silent"
            ShellExecute $script:Settings.ServerPath $local:ArgList $true -Stack $($StackLevel+1)
            Output-Result $true -Indent $($StackLevel+1)
        }

        $local:ServiceName = DCOM_GetAppServiceName $script:ServerAppID $($StackLevel+1)
        if ( $local:ServiceName -ne $null )
        {
            Output-Write "• Настроить службу `"$local:ServiceName`" на вход в систему с учетной записью: `"$($script:Settings.ServerUser.Name)`"" -Indent $StackLevel

            try 
            {
                $local:BSTR = [System.Runtime.InteropServices.Marshal]::SecureStringToBSTR($script:Settings.ServerUser.Password)
                $local:Password = [System.Runtime.InteropServices.Marshal]::PtrToStringAuto($local:BSTR)

                # Uninstall 'Tekon OPC Server'            
                $local:ArgList = "-s -u --stop --wait --silent"
                ShellExecute $script:Settings.ServerPath $local:ArgList $true -Stack $($StackLevel+1)

                # Install 'Tekon OPC Server'
                $local:ArgList = "-s -i --wait --silent"
                $local:ArgList += " --user " + $script:Settings.ServerUser.Name
                $local:ArgList += " --password " + $local:Password

                ShellExecute $script:Settings.ServerPath $local:ArgList $true -Stack $($StackLevel+1)

                Output-Result $true -Indent $($StackLevel+1)
            }
            catch
            {
                Output-Result $false -Indent $($StackLevel+1)
            }
        }
        else
        {
            Output-Warning "Не удалось получить имя службы `"AppID:$($script:ServerAppID)`"" -Indent 1
        }        
        
        ProgressBarInc       
    }




    # ScadaUser::GrantRights to control Tekon OPC Server
    if ($script:Settings.ScadaUser.ServerControl)
    {
        $local:AppName = DCOM_GetAppServiceName $script:ServerAppID $($StackLevel+1)
        if ( $local:AppName -ne $null )
        {
            Output-Write "• Предоставить пользователю `"$($script:Settings.ScadaUser.Name)`" права на управление службой `"$local:AppName`"" -Indent $StackLevel

            $local:UserSID =  UserMgt_GetSID $script:Settings.ScadaUser.Name $($StackLevel+1)
            if ($local:UserSID -ne $null)
            {
                $local:result = SvcUtil_GrantServiceControlRightsByAppID $script:ServerAppID $local:UserSID $($StackLevel+1) 
                Output-Result $local:result -Indent $($StackLevel+1)
            }
            else
            {
                Output-Warning "Не удалось получить SID пользователя `"$($script:Settings.ScadaUser.Name)`"" -Indent $StackLevel
            }            
        }
        else
        {
             Output-Warning "Не удалось получить имя службы `"AppID:$($script:ServerAppID)`"" -Indent 1
        }
                
        ProgressBarInc 
    }


    # Grant ScadaUser access rights to the folder
    if ($script:Settings.ScadaUser.ServerFolder)
    {
        $local:Folder = split-path -Parent $script:Settings.ServerPath
        if ( $local:Folder -ne $null )
        {
            Output-Write "• Предоставить пользователю `"$($script:Settings.ScadaUser.Name)`" разрешения на полный доступ к каталогу  `"$local:Folder`"" -Indent $StackLevel

            $local:result = FS_GrantAccessRightsToFolder $local:Folder $script:Settings.ScadaUser.Name $($StackLevel+1)
            Output-Result $local:result -Indent $($StackLevel+1)
        }

        ProgressBarInc 
    }


    # ScadaUser::GrantRights to control OPC Enum
    if ($script:Settings.ScadaUser.OpcEnumControl)
    {
        $local:AppName = DCOM_GetAppServiceName $script:OpcEnumAppID $($StackLevel+1)
        if ( $local:AppName -ne $null )
        {
            Output-Write "• Предоставить пользователю `"$($script:Settings.ScadaUser.Name)`" права на управление службой `"$local:AppName`"" -Indent $StackLevel

            $local:UserSID =  UserMgt_GetSID $script:Settings.ScadaUser.Name $($StackLevel+1)
            if ($local:UserSID -ne $null)
            {
                $local:result = SvcUtil_GrantServiceControlRightsByAppID $script:OpcEnumAppID $local:UserSID $($StackLevel+1)
                Output-Result $local:result -Indent $($StackLevel+1)
            }
            else
            {
                Output-Warning "Не удалось получить SID пользователя `"$($script:Settings.ScadaUser.Name)`"" -Indent $StackLevel
            }            
        }
        else
        {
             Output-Warning "Не удалось получить имя службы `"AppID:$($script:ServerAppID)`"" -Indent 1
        }
                
        ProgressBarInc
    }


    # Setting up access rights via DСOM for the "ScadaUser" user
    if ($script:Settings.ScadaUser.AllowDCOM)
    {
        Output-Write "• Предоставить пользователю `"$($script:Settings.ScadaUser.Name)`" права на доступ к DCOM компонентам на этом компьютере" -Indent $StackLevel

        $local:result1 = $(DCOM_ResetSettingsToDefault $($StackLevel+1))
        $local:result2 = $(DCOM_AddToDefaults $script:Settings.ScadaUser.Name $($StackLevel+1))
        $local:result3 = $(DCOM_AddToLimits   $script:Settings.ScadaUser.Name $($StackLevel+1))

        Output-Result $($local:result1 -and $local:result2 -and $local:result3) -Indent $($StackLevel+1)

        ProgressBarInc
    }


    # Grant "ScadaUser" user access rights to the DCOM application "Tekon OPC Server"
    if ($script:Settings.ScadaUser.AllowServer)
    {
        $local:AppName = DCOM_GetAppName $script:ServerAppID $($StackLevel+1)
        if ($local:AppName -ne $null)
        {
            Output-Write "• Предоставить пользователю `"$($script:Settings.ScadaUser.Name)`" права на доступ к DCOM приложению `"$local:AppName`"" -Indent $StackLevel

            $local:result = DCOM_AddUserPermissionsToApp $script:ServerAppID $script:Settings.ScadaUser.Name $($StackLevel+1)

            Output-Result $local:result -Indent $($StackLevel+1)
        }
        else
        {
            Output-Warning "Не удалось получить имя DCOM приложения `"AppID:$($script:ServerAppID)`"" -Indent 1
        }

        ProgressBarInc
    }


    # Grant "ScadaUser" user access rights to the DCOM application "OPC Enum"
    if ($script:Settings.ScadaUser.AllowOpcEnum)
    {
        $local:AppName = DCOM_GetAppName $script:OpcEnumAppID $($StackLevel+1)
        if ($local:AppName -ne $null)
        {
            Output-Write "• Предоставить пользователю `"$($script:Settings.ScadaUser.Name)`" права на доступ к DCOM приложению `"$local:AppName`"" -Indent $StackLevel

            $local:result = DCOM_AddUserPermissionsToApp $script:OpcEnumAppID $script:Settings.ScadaUser.Name $($StackLevel+1)

            Output-Result $local:result -Indent $($StackLevel+1)
        }
        else
        {
            Output-Warning "Не удалось получить имя DCOM приложения `"AppID:$($script:OpcEnumAppID)`"" -Indent 1
        }

        ProgressBarInc
    }



    # DCOMUsers::GrantRights to control Tekon OPC Server
    if ($script:Settings.ScadaUser.ServerControl)
    {
        $local:GroupName = UserMgt_GetName $sidDCOMUsers $($StackLevel+1)
        if ($local:GroupName -ne $null)
        {
            $local:AppName = DCOM_GetAppServiceName $script:ServerAppID $($StackLevel+1)
            if ( $local:AppName -ne $null )
            {
                Output-Write "• Предоставить группе `"$($local:GroupName)`" права на управление службой `"$local:AppName`"" -Indent $StackLevel

                $local:result = SvcUtil_GrantServiceControlRightsByAppID $script:ServerAppID $sidDCOMUsers $($StackLevel+1) 
                Output-Result $local:result -Indent $($StackLevel+1)          
            }
            else
            {
                Output-Warning "Не удалось получить имя службы `"AppID:$($script:ServerAppID)`"" -Indent 1
            }       
        }
        else
        {
            Output-Warning "Не удалось получить имя группы `"sid:$sidDCOMUsers`"" -Indent 1
        }
                
        ProgressBarInc 
    }


    # DCOMUsers::GrantRights to control OPC Enum
    if ($script:Settings.ScadaUser.ServerControl)
    {
        $local:GroupName = UserMgt_GetName $sidDCOMUsers $($StackLevel+1)
        if ($local:GroupName -ne $null)
        {
            $local:AppName = DCOM_GetAppServiceName $script:OpcEnumAppID $($StackLevel+1)
            if ( $local:AppName -ne $null )
            {
                Output-Write "• Предоставить группе `"$($local:GroupName)`" права на управление службой `"$local:AppName`"" -Indent $StackLevel

                $local:result = SvcUtil_GrantServiceControlRightsByAppID $script:OpcEnumAppID $sidDCOMUsers $($StackLevel+1) 
                Output-Result $local:result -Indent $($StackLevel+1)          
            }
            else
            {
                 Output-Warning "Не удалось получить имя службы `"AppID:$($script:OpcEnumAppID)`"" -Indent 1
            }
        }
        else
        {
            Output-Warning "Не удалось получить имя группы `"sid:$sidDCOMUsers`"" -Indent 1
        }
                
        ProgressBarInc 
    }


    # Setting up access rights via DСOM for the "DCOM Users" group
    if ($script:Settings.DCOMUsers.AllowDCOM)
    {
        $local:GroupName = UserMgt_GetName $sidDCOMUsers $($StackLevel+1)
        if ($local:GroupName -ne $null)
        {
            Output-Write "• Предоставить группе `"$($local:GroupName)`" права на доступ к DCOM компонентам на этом компьютере" -Indent $StackLevel

            $local:result1 = $(DCOM_ResetSettingsToDefault $($StackLevel+1))
            $local:result2 = $(DCOM_AddToDefaultsBySID $sidDCOMUsers $($StackLevel+1))
            $local:result3 = $(DCOM_AddToLimitsBySID   $sidDCOMUsers $($StackLevel+1))

            Output-Result $($local:result1 -and $local:result2 -and $local:result3) -Indent $($StackLevel+1)
        }
        else
        {
            Output-Warning "Не удалось получить имя группы `"sid:$sidDCOMUsers`"" -Indent 1
        }

        ProgressBarInc
    }


    # Grant "DCOM Users" group access rights to the DCOM application "Tekon OPC Server"
    if ($script:Settings.DCOMUsers.AllowServer)
    {
        $local:GroupName = UserMgt_GetName $sidDCOMUsers $($StackLevel+1)
        if ($local:GroupName -ne $null)
        {
            $local:AppName = DCOM_GetAppName $script:ServerAppID $($StackLevel+1)
            if ($local:AppName -ne $null)
            {
                Output-Write "• Предоставить группе `"$($local:GroupName)`" права на доступ к DCOM приложению `"$local:AppName`"" -Indent $StackLevel

                $local:result = DCOM_AddUserPermissionsToApp $script:ServerAppID $local:GroupName $($StackLevel+1)

                Output-Result $local:result -Indent $($StackLevel+1)
            }
            else
            {
                Output-Warning "Не удалось получить имя DCOM приложения `"AppID:$($script:ServerAppID)`"" -Indent 1
            }
        }
        else
        {
            Output-Warning "Не удалось получить имя группы `"sid:$sidDCOMUsers`"" -Indent 1
        }

        ProgressBarInc
    }


    # Grant "DCOM Users" group access rights to the DCOM application "OPC Enum"
    if ($script:Settings.DCOMUsers.AllowOpcEnum)
    {
        $local:GroupName = UserMgt_GetName $sidDCOMUsers $($StackLevel+1)
        if ($local:GroupName -ne $null)
        {
            $local:AppName = DCOM_GetAppName $script:OpcEnumAppID $($StackLevel+1)
            if ($local:AppName -ne $null)
            {
                Output-Write "• Предоставить группе `"$($local:GroupName)`" права на доступ к DCOM приложению `"$local:AppName`"" -Indent $StackLevel

                $local:result = DCOM_AddUserPermissionsToApp $script:OpcEnumAppID $local:GroupName $($StackLevel+1)

                Output-Result $local:result -Indent $($StackLevel+1)
            }
            else
            {
                Output-Warning "Не удалось получить имя DCOM приложения `"AppID:$($script:OpcEnumAppID)`"" -Indent 1
            }
        }
        else
        {
            Output-Warning "Не удалось получить имя группы `"sid:$sidDCOMUsers`"" -Indent 1
        }

        ProgressBarInc
    }
    

    ### Configuring the Firewall
    Output-Write "• Настрока Брандмауэра:" -Indent 1 -LinesBefore 1

    Output-Write "• Разрешить входящий трафик ICMPv4 (Ping)" -Indent $StackLevel
    $local:RuleName = "Tekon. Allow ICMPv4 (Inbound)"
    $local:result = Update-NetFirewallRule -DisplayName $local:RuleName -Direction Inbound -Action Allow -Protocol ICMPv4 -Stack $($StackLevel+1)
    Output-Result $local:result -Indent $($StackLevel+1)
    ProgressBarInc

    Output-Write "• Разрешить входящий трафик TCP на порт 135 (DCOM)" -Indent $StackLevel
    $local:RuleName = "Tekon. Allow TCP:135 (Inbound)"
    $local:result = Update-NetFirewallRule -DisplayName $local:RuleName -Direction Inbound -Action Allow -Protocol TCP -LocalPort 135 -Stack $($StackLevel+1)
    Output-Result $local:result -Indent $($StackLevel+1)
    ProgressBarInc    
    
    if ( $script:Settings.ScadaPath -ne $null )
    {
        Output-Write "• Разрешить `"$(Split-Path $script:Settings.ScadaPath -Leaf)`" входяшие подключения" -Indent $StackLevel
        $RuleName = "Tekon. Allow Scada (Inbound)"
        $result = Update-NetFirewallRule -DisplayName $RuleName -Direction Inbound -Action Allow -Program $script:Settings.ScadaPath -Stack $($StackLevel+1)
        Output-Result $local:result -Indent $($StackLevel+1)
        ProgressBarInc

        Output-Write "• Разрешить `"$(Split-Path $script:Settings.ScadaPath -Leaf)`" исходящие подключения" -Indent $StackLevel
        $RuleName = "Tekon. Allow Scada (Outbound)"
        $result = Update-NetFirewallRule -DisplayName $RuleName -Direction Outbound -Action Allow -Program $script:Settings.ScadaPath -Stack $($StackLevel+1)
        Output-Result $local:result -Indent $($StackLevel+1)
        ProgressBarInc
    }

    if ( $script:Settings.ServerPath -ne $null )
    {
        Output-Write "• Разрешить `"$(Split-Path $script:Settings.ServerPath -Leaf)`" входяшие подключения" -Indent $StackLevel
        $RuleName = "Tekon. Allow Server (Inbound)"
        $result = Update-NetFirewallRule -DisplayName $RuleName -Direction Inbound -Action Allow -Program $script:Settings.ServerPath -Stack $($StackLevel+1)
        Output-Result $local:result -Indent $($StackLevel+1)
        ProgressBarInc

        Output-Write "• Разрешить `"$(Split-Path $script:Settings.ServerPath -Leaf)`" исходящие подключения" -Indent $StackLevel
        $RuleName = "Tekon. Allow Server (Outbound)"
        $result = Update-NetFirewallRule -DisplayName $RuleName -Direction Outbound -Action Allow -Program $script:Settings.ServerPath -Stack $($StackLevel+1)
        Output-Result $local:result -Indent $($StackLevel+1)
        ProgressBarInc

        $local:AppName = DCOM_GetAppServiceName $script:ServerAppID $($StackLevel+1)
        if ( $local:AppName -ne $null )
        {
            Output-Write "• Остановить службу `"$($local:AppName)`"" -Indent $StackLevel

            net stop "$local:AppName" | out-null

            $local:result = SvcUtil_GrantServiceControlRightsByAppID $script:ServerAppID $sidDCOMUsers $($StackLevel+1) 
            Output-Result $local:result -Indent $($StackLevel+1)          
        }
        else
        {
            Output-Warning "Не удалось получить имя службы `"AppID:$($script:OpcEnumAppID)`"" -Indent 1
        }
    }

    if ( $script:Settings.OpcEnumPath -ne $null )
    {
        Output-Write "• Разрешить `"$(Split-Path $script:Settings.OpcEnumPath -Leaf)`" входяшие подключения" -Indent $StackLevel
        $RuleName = "Tekon. Allow OpcEnum (Inbound)"
        $result = Update-NetFirewallRule -DisplayName $RuleName -Direction Inbound -Action Allow -Program $script:Settings.OpcEnumPath -Stack $($StackLevel+1)
        Output-Result $local:result -Indent $($StackLevel+1)
        ProgressBarInc

        $local:AppName = DCOM_GetAppServiceName $script:OpcEnumAppID $($StackLevel+1)
        if ( $local:AppName -ne $null )
        {
            Output-Write "• Остановить службу `"$($local:AppName)`"" -Indent $StackLevel

            net stop "$local:AppName" | out-null

            $local:result = SvcUtil_GrantServiceControlRightsByAppID $script:OpcEnumAppID $sidDCOMUsers $($StackLevel+1) 
            Output-Result $local:result -Indent $($StackLevel+1)          
        }
        else
        {
            Output-Warning "Не удалось получить имя службы `"AppID:$($script:OpcEnumAppID)`"" -Indent 1
        }
    }

    $ProgressBar.Value = 100
    Output-Write "Выполнено"
}


##########################################################################
## Create GUI
function ShowMainForm
{ 
    $main_form = New-Object System.Windows.Forms.Form
    $main_form.Text = "АСУД Scada. DCOM Access Assistant"
    $main_form.Size = New-Object System.Drawing.Size(750, 750)    
    $main_form.FormBorderStyle = [System.Windows.Forms.FormBorderStyle]::Fixed3D
    $main_form.SizeGripStyle = [System.Windows.Forms.SizeGripStyle]::Hide
    $main_form.MaximizeBox = $false
    $main_form.MinimizeBox = $false
    $main_form.StartPosition = 'CenterScreen'

    $label_desc = New-Object System.Windows.Forms.Label
    $label_desc.Location = New-Object System.Drawing.Point(27, 17)
    $label_desc.Size = New-Object System.Drawing.Size(700, 45)
    $label_desc.Text = "Данный мастер создает необходимых для работы системы Пользователей, а так же настраивает: `nБрандмауэр, Службы Компонентов DCOM и Локальную политику безопасности"
    $main_form.Controls.Add($label_desc)

    $gb_soft = New-Object System.Windows.Forms.GroupBox
    $gb_soft.Location = New-Object System.Drawing.Point(10, 67)
    $gb_soft.Size = New-Object System.Drawing.Size(720, 70)
    $gb_soft.text = "Что настраиваем?"
    $main_form.Controls.Add($gb_soft)    
       
    $cb_server = New-Object System.Windows.Forms.CheckBox
    $cb_server.location = New-Object System.Drawing.Point(17, 27)
    $cb_server.Size = New-Object System.Drawing.Size(130, 15)
    $cb_server.UseVisualStyleBackColor = $True
    $cb_server.Text = "Текон OPC сервер"
    $cb_server.TabIndex = 0
    $cb_server.checkAlign = "MiddleLeft"    
    $cb_server.DataBindings.DefaultDataSourceUpdateMode = 0
    $cb_server.Name = "Server"
    $cb_server.Checked = $false
    $gb_soft.Controls.Add($cb_server)

    $cb_scada = New-Object System.Windows.Forms.CheckBox
    $cb_scada.location = New-Object System.Drawing.Point(17, 47)
    $cb_scada.Size = New-Object System.Drawing.Size(130, 15)
    $cb_scada.UseVisualStyleBackColor = $True
    $cb_scada.Text = "АСУД Scada"
    $cb_scada.TabIndex = 1
    $cb_scada.checkAlign = "MiddleLeft"    
    $cb_scada.DataBindings.DefaultDataSourceUpdateMode = 0
    $cb_scada.Name = "Scada"
    $cb_scada.Checked = $false
    $gb_soft.Controls.Add($cb_scada)       
    
    # parameter group
    $gb_param = New-Object System.Windows.Forms.GroupBox
    $gb_param.Location = New-Object System.Drawing.Point(10, 147)
    $gb_param.Size = New-Object System.Drawing.Size(720, 250)
    $gb_param.Text = "Параметры"  
    $main_form.Controls.Add($gb_param)

    # server exe
    $ltb_server_exe = New-Object System.Windows.Forms.Label
    $ltb_server_exe.Enabled = $false
    $ltb_server_exe.Location = New-Object System.Drawing.Point(17, 29)
    $ltb_server_exe.Size = New-Object System.Drawing.Size(190, 15)    
    $ltb_server_exe.Text = "Исполняемый файл OPC сервера:"
    $ltb_server_exe.Enabled = $false
    $gb_param.Controls.Add($ltb_server_exe)

    $tb_server_exe = New-Object System.Windows.Forms.TextBox
    $tb_server_exe.Enabled = $false
    $tb_server_exe.Location = New-Object System.Drawing.Point(207, 27)
    $tb_server_exe.Size = New-Object System.Drawing.Size(467, 15)
    $tb_server_exe.Text = ""
    $tb_server_exe.TabIndex = 2
    $gb_param.Controls.Add($tb_server_exe)

    $btn_server_exe = New-Object System.Windows.Forms.Button
    $btn_server_exe.Enabled = $false
    $btn_server_exe.Location = New-Object System.Drawing.Point(675, 27)
    $btn_server_exe.Size = New-Object System.Drawing.Size(30,20)
    $btn_server_exe.TabIndex = 3
    $btn_server_exe.Text = '...'        
    $gb_param.Controls.Add($btn_server_exe)
    $btn_server_exe.Add_Click(
    {
        $FileBrowser = New-Object System.Windows.Forms.OpenFileDialog -Property @{ 
            InitialDirectory = $(Split-Path -Path $tb_server_exe.Text)
            RestoreDirectory = $true
            Filter = 'opc server (opcsrv*.exe)|opcsrv*.exe|All files (*.*)|*.*'
        }

        if ($FileBrowser.ShowDialog() -eq [System.Windows.Forms.DialogResult]::OK)
        {
            $tb_server_exe.Text = $FileBrowser.FileName
        }    

        $FileBrowser.Dispose()
    })

    # scada exe
    $ltb_scada_exe = New-Object System.Windows.Forms.Label
    $ltb_scada_exe.Enabled = $false
    $ltb_scada_exe.Location = New-Object System.Drawing.Point(17, 59)
    $ltb_scada_exe.Size = New-Object System.Drawing.Size(190, 15)    
    $ltb_scada_exe.Text = "Исполняемый файл АСУД Scada:"
    $ltb_scada_exe.Enabled = $false
    $gb_param.Controls.Add($ltb_scada_exe)

    $tb_scada_exe = New-Object System.Windows.Forms.TextBox
    $tb_scada_exe.Enabled = $false
    $tb_scada_exe.Location = New-Object System.Drawing.Point(207, 57)
    $tb_scada_exe.Size = New-Object System.Drawing.Size(467, 15)
    $tb_scada_exe.Text = ""
    $tb_scada_exe.TabIndex = 4
    $gb_param.Controls.Add($tb_scada_exe)

    $btn_scada_exe = New-Object System.Windows.Forms.Button
    $btn_scada_exe.Enabled = $false
    $btn_scada_exe.Location = New-Object System.Drawing.Point(675, 57)
    $btn_scada_exe.Size = New-Object System.Drawing.Size(30,20)
    $btn_scada_exe.TabIndex = 5
    $btn_scada_exe.Text = '...'        
    $gb_param.Controls.Add($btn_scada_exe)
    $btn_scada_exe.Add_Click(
    {
        $FileBrowser = New-Object System.Windows.Forms.OpenFileDialog -Property @{ 
            InitialDirectory = $(Split-Path -Path $tb_scada_exe.Text)
            RestoreDirectory = $true
            Filter = 'asud scada (scada*.exe)|scada*.exe|All files (*.*)|*.*'
        }

        if ($FileBrowser.ShowDialog() -eq [System.Windows.Forms.DialogResult]::OK)
        {
            $tb_scada_exe.Text = $FileBrowser.FileName
        }    

        $FileBrowser.Dispose()    
    })    


    # user server
    $gb_server_user = New-Object System.Windows.Forms.GroupBox
    $gb_server_user.Enabled = $false
    $gb_server_user.Location = New-Object System.Drawing.Point(17, 87)
    $gb_server_user.Size = New-Object System.Drawing.Size(320, 150)
    $gb_server_user.Text = "Пользователь Server"  
    $gb_param.Controls.Add($gb_server_user)

    $ltb_server_usr = New-Object System.Windows.Forms.Label
    $ltb_server_usr.Location = New-Object System.Drawing.Point(17, 27)
    $ltb_server_usr.Size     = New-Object System.Drawing.Size(50, 15)    
    $ltb_server_usr.Text     = "Логин:"
    $ltb_server_usr.Enabled  = $false
    $gb_server_user.Controls.Add($ltb_server_usr)

    $tb_server_usr = New-Object System.Windows.Forms.TextBox
    $tb_server_usr.MaxLength = 20
    $tb_server_usr.Location  = New-Object System.Drawing.Point(67, 27)
    $tb_server_usr.Size      = New-Object System.Drawing.Size(210, 15)
    $tb_server_usr.Text      = "tekon.server"
    $tb_server_usr.TabIndex  = 6
    $gb_server_user.Controls.Add($tb_server_usr)

    $btn_server_usr = New-Object System.Windows.Forms.Button
    $btn_server_usr.Location = New-Object System.Drawing.Point(280, 27)
    $btn_server_usr.Size     = New-Object System.Drawing.Size(25,20)
    $btn_server_usr.TabIndex = 7
    $btn_server_usr.Text     = '...'        
    $gb_server_user.Controls.Add($btn_server_usr)
    $btn_server_usr.Add_Click(
    {
        $cred = Get-Credential -credential $tb_server_usr.Text
        if ($cred -ne $null)
        {
            $tb_server_usr.Text = $cred.GetNetworkCredential().UserName
            $tb_server_passwd.Text = $cred.GetNetworkCredential().password
        }
    })
     
    $ltb_server_passwd = New-Object System.Windows.Forms.Label
    $ltb_server_passwd.Location = New-Object System.Drawing.Point(17, 57)
    $ltb_server_passwd.Size     = New-Object System.Drawing.Size(50, 15)
    $ltb_server_passwd.Text     = "Пароль:"
    $ltb_server_passwd.Enabled  = $false
    $gb_server_user.Controls.Add($ltb_server_passwd)
    
    $tb_server_passwd = New-Object System.Windows.Forms.MaskedTextBox
    $tb_server_passwd.PasswordChar = '*'
    $tb_server_passwd.Location     = New-Object System.Drawing.Point(67, 57)
    $tb_server_passwd.Size         = New-Object System.Drawing.Size(210, 15)
    $tb_server_passwd.Text         = ""
    $tb_server_passwd.TabIndex = 8
    $gb_server_user.Controls.Add($tb_server_passwd)   

    $cb_server_localdeny = New-Object System.Windows.Forms.CheckBox
    $cb_server_localdeny.location = New-Object System.Drawing.Point(17, 87)
    $cb_server_localdeny.Size = New-Object System.Drawing.Size(170, 15)
    $cb_server_localdeny.UseVisualStyleBackColor = $True
    $cb_server_localdeny.Text = "Запретить локальный вход"
    $cb_server_localdeny.TabIndex = 9
    $cb_server_localdeny.checkAlign = "MiddleLeft"    
    $cb_server_localdeny.DataBindings.DefaultDataSourceUpdateMode = 0
    $cb_server_localdeny.Name = "ServerLocalDeny"
    $cb_server_localdeny.Checked = $false
    $gb_server_user.Controls.Add($cb_server_localdeny)

    $cb_server_networkdeny = New-Object System.Windows.Forms.CheckBox
    $cb_server_networkdeny.location = New-Object System.Drawing.Point(17, 107)
    $cb_server_networkdeny.Size = New-Object System.Drawing.Size(170, 15)
    $cb_server_networkdeny.UseVisualStyleBackColor = $True
    $cb_server_networkdeny.Text = "Отказать в доступе из сети"
    $cb_server_networkdeny.TabIndex = 10
    $cb_server_networkdeny.checkAlign = "MiddleLeft"    
    $cb_server_networkdeny.DataBindings.DefaultDataSourceUpdateMode = 0
    $cb_server_networkdeny.Name = "ServerNetworkDeny"
    $cb_server_networkdeny.Checked = $false
    $gb_server_user.Controls.Add($cb_server_networkdeny)


    $cb_server_networkallow = New-Object System.Windows.Forms.CheckBox
    $cb_server_networkallow.Enabled = $false
    $cb_server_networkallow.location = New-Object System.Drawing.Point(17, 127)
    $cb_server_networkallow.Size = New-Object System.Drawing.Size(170, 15)
    $cb_server_networkallow.UseVisualStyleBackColor = $True
    $cb_server_networkallow.Text = "Разрешить доступ из сети"
    $cb_server_networkallow.TabIndex = 15
    $cb_server_networkallow.checkAlign = "MiddleLeft"    
    $cb_server_networkallow.DataBindings.DefaultDataSourceUpdateMode = 0
    $cb_server_networkallow.Name = "ServerNetworkDeny"
    $cb_server_networkallow.Checked = $false
    $gb_server_user.Controls.Add($cb_server_networkallow)

    # user scada 
    $gb_scada_user = New-Object System.Windows.Forms.GroupBox
    $gb_scada_user.Enabled = $false
    $gb_scada_user.Location = New-Object System.Drawing.Point(354, 87)
    $gb_scada_user.Size = New-Object System.Drawing.Size(320, 150)
    $gb_scada_user.Text = "Пользователь Scada"  
    $gb_param.Controls.Add($gb_scada_user)


    $ltb_scada_usr = New-Object System.Windows.Forms.Label
    $ltb_scada_usr.Location = New-Object System.Drawing.Point(17, 27)
    $ltb_scada_usr.Size = New-Object System.Drawing.Size(50, 15)    
    $ltb_scada_usr.Text = "Логин:"
    $ltb_scada_usr.Enabled = $false
    $gb_scada_user.Controls.Add($ltb_scada_usr)

    $tb_scada_usr = New-Object System.Windows.Forms.TextBox
    $tb_scada_usr.MaxLength = 20
    $tb_scada_usr.Location = New-Object System.Drawing.Point(67, 27)
    $tb_scada_usr.Size = New-Object System.Drawing.Size(210, 15)
    $tb_scada_usr.Text = "tekon.scada"
    $tb_scada_usr.TabIndex = 11
    $gb_scada_user.Controls.Add($tb_scada_usr)

    $btn_scada_usr = New-Object System.Windows.Forms.Button
    $btn_scada_usr.Location = New-Object System.Drawing.Point(280, 27)
    $btn_scada_usr.Size = New-Object System.Drawing.Size(25,20)
    $btn_scada_usr.TabIndex = 12
    $btn_scada_usr.Text = '...'        
    $gb_scada_user.Controls.Add($btn_scada_usr)
    $btn_scada_usr.Add_Click(
    {
        $cred = Get-Credential -credential $tb_scada_usr.Text
        if ($cred -ne $null)
        {
            $tb_scada_usr.Text = $cred.GetNetworkCredential().UserName
            $tb_scada_passwd.Text = $cred.GetNetworkCredential().password
        }
    })
     
    $ltb_scada_passwd = New-Object System.Windows.Forms.Label
    $ltb_scada_passwd.Location = New-Object System.Drawing.Point(17, 57)
    $ltb_scada_passwd.Size = New-Object System.Drawing.Size(50, 15)
    $ltb_scada_passwd.Text = "Пароль:"
    $ltb_scada_passwd.Enabled = $false
    $gb_scada_user.Controls.Add($ltb_scada_passwd)
    
    $tb_scada_passwd = New-Object System.Windows.Forms.MaskedTextBox
    $tb_scada_passwd.PasswordChar = '*'
    $tb_scada_passwd.Location = New-Object System.Drawing.Point(67, 57)
    $tb_scada_passwd.Size = New-Object System.Drawing.Size(210, 15)
    $tb_scada_passwd.Text = ""
    $tb_scada_passwd.TabIndex = 13
    $gb_scada_user.Controls.Add($tb_scada_passwd)    

    $cb_scada_localdeny = New-Object System.Windows.Forms.CheckBox
    $cb_scada_localdeny.location = New-Object System.Drawing.Point(17, 87)
    $cb_scada_localdeny.Size = New-Object System.Drawing.Size(170, 15)
    $cb_scada_localdeny.UseVisualStyleBackColor = $True
    $cb_scada_localdeny.Text = "Запретить локальный вход"
    $cb_scada_localdeny.TabIndex = 14
    $cb_scada_localdeny.checkAlign = "MiddleLeft"    
    $cb_scada_localdeny.DataBindings.DefaultDataSourceUpdateMode = 0
    $cb_scada_localdeny.Name = "ServerLocalDeny"
    $cb_scada_localdeny.Checked = $false
    $gb_scada_user.Controls.Add($cb_scada_localdeny)

    $cb_scada_networkallow = New-Object System.Windows.Forms.CheckBox
    $cb_scada_networkallow.Enabled = $false
    $cb_scada_networkallow.location = New-Object System.Drawing.Point(17, 107)
    $cb_scada_networkallow.Size = New-Object System.Drawing.Size(170, 15)
    $cb_scada_networkallow.UseVisualStyleBackColor = $True
    $cb_scada_networkallow.Text = "Разрешить доступ из сети"
    $cb_scada_networkallow.TabIndex = 15
    $cb_scada_networkallow.checkAlign = "MiddleLeft"    
    $cb_scada_networkallow.DataBindings.DefaultDataSourceUpdateMode = 0
    $cb_scada_networkallow.Name = "ScadaNetworkDeny"
    $cb_scada_networkallow.Checked = $false
    $gb_scada_user.Controls.Add($cb_scada_networkallow)

    $OnSelectSoftware = 
    {
        if ($cb_server.Checked)
        {
            $ltb_server_exe.Enabled = $true
            $tb_server_exe.Enabled = $true
            $btn_server_exe.Enabled = $true

            $gb_server_user.Enabled = $true

            $gb_scada_user.Enabled = $true

            $cb_scada_networkallow.Checked= $true
        }
        else
        {
            $ltb_server_exe.Enabled = $false
            $tb_server_exe.Enabled = $false
            $btn_server_exe.Enabled = $false

            $gb_scada_user.Enabled = $false

            if ($cb_scada.Checked)
            {
                $gb_server_user.Enabled = $true                
            }
            else
            {
                $gb_server_user.Enabled = $false
            }

            $cb_scada_networkallow.Checked= $false
        }

        if ($cb_scada.Checked)
        {
            $ltb_scada_exe.Enabled = $true
            $tb_scada_exe.Enabled  = $true
            $btn_scada_exe.Enabled = $true
        
            $cb_server_networkdeny.Enabled  = $false
            $cb_server_networkdeny.Checked  = $false
            $cb_server_networkallow.Checked = $true
        }
        else
        {
            $ltb_scada_exe.Enabled          = $false
            $tb_scada_exe.Enabled           = $false
            $btn_scada_exe.Enabled          = $false
            
            $cb_server_networkdeny.Enabled  = $true
            $cb_server_networkallow.Checked = $false
        }
    }

    $cb_server.Add_CheckStateChanged($OnSelectSoftware)
    $cb_scada.Add_CheckStateChanged($OnSelectSoftware)

    $tooltip = New-Object System.Windows.Forms.ToolTip
    $tooltip.SetToolTip( $cb_scada,         "АСУД Scada будет работать на текущем компьютере")
    $tooltip.SetToolTip( $cb_server,        "Служба Tekon OPC Server зарегистрирована на текущем компьютере")
    $tooltip.SetToolTip( $ltb_scada_usr,    "Пользователь, используемый для подключения к Tekon OPC Server")
    $tooltip.SetToolTip( $tb_scada_usr,     "Пользователь, используемый для подключения к Tekon OPC Server")
    $tooltip.SetToolTip( $ltb_server_usr,   "Пользователь, от имени которого запускается служба Tekon OPC Server")
    $tooltip.SetToolTip( $tb_server_usr,    "Пользователь, от имени которого запускается служба Tekon OPC Server")


    $btn_ok = New-Object System.Windows.Forms.Button
    $btn_ok.Location = New-Object System.Drawing.Point(565, 405)
    $btn_ok.Size = New-Object System.Drawing.Size(75,23)
    $btn_ok.TabIndex = 16
    $btn_ok.Text = "Проверить"
    $btn_ok.Add_Click(
    {
        if ($btn_ok.Text -eq "Проверить")
        {
            $main_form.ControlBox = $false
            $gb_soft.Enabled      = $false
            $gb_param.Enabled     = $false

            $cb_debug.Enabled     = $false
            $btn_ok.Enabled       = $false
            $btn_cancel.Enabled   = $false
            $btn_clear.Enabled    = $false

            if (-not (AnalyzeParameters))
            {
                $ProgressBar.Visible = $false

                Output-Result $false

                $main_form.ControlBox = $true    
                $gb_soft.Enabled      = $true
                $gb_param.Enabled     = $true

                $cb_debug.Enabled     = $true
                $btn_ok.Enabled       = $true
                $btn_cancel.Enabled   = $true
                $btn_clear.Enabled    = $true
                
                if ($script:SelectedWnd -ne $null)
                {
                    $script:SelectedWnd.Select()
                    $script:SelectedWnd.SelectAll()
                }

                return
            }

            $ProgressBar.Visible = $false

            $cb_debug.Enabled   = $true
            $btn_ok.Enabled     = $true
            $btn_cancel.Enabled = $true
            $btn_clear.Enabled  = $true

            $btn_ok.Text = "Применить"
            $btn_cancel.Text = "Отмена"
        }
        elseif ($btn_ok.Text -eq "Применить")
        {
            $cb_debug.Enabled      = $false
            $btn_ok.Enabled        = $false
            $btn_cancel.Enabled    = $false
            $btn_clear.Enabled     = $false

            ApplySettings

#            $ProgressBar.Visible  = $false

            $main_form.ControlBox = $true    
            $gb_soft.Enabled      = $true
            $gb_param.Enabled     = $true

            $cb_debug.Enabled     = $true
            $btn_ok.Enabled       = $true
            $btn_cancel.Enabled   = $true
            $btn_clear.Enabled    = $true

            $btn_ok.Text          = "Проверить"            
            $btn_cancel.Text      = "Закрыть"
        }
        else
        {
            [System.Environment]::Exit(1)
        }
    })
    $main_form.Controls.Add($btn_ok)

    $btn_cancel = New-Object System.Windows.Forms.Button
    $btn_cancel.Location = New-Object System.Drawing.Point(650, 405)
    $btn_cancel.Size = New-Object System.Drawing.Size(75,23)
    $btn_cancel.TabIndex = 17
    $btn_cancel.Text = "Закрыть"
    $btn_cancel.Add_Click(
    {
        if ($btn_ok.Text -eq "Проверить")
        {
            $main_form.Close()
        }
        elseif ($btn_ok.Text -eq "Применить")
        {
            $main_form.ControlBox = $true
            $gb_soft.Enabled      = $true
            $gb_param.Enabled     = $true
            $btn_ok.Enabled       = $true
            $btn_cancel.Enabled   = $true

            $btn_ok.Text          = "Проверить"
            $btn_cancel.Text      = "Закрыть"

            # Clear output if not in debug mode (-not IgnoreOnlyDebug)
            if ( -not $cb_debug.Checked ) { Output-Clear } else { Output-Write " " }

            Output-Write "● Необходимо заполнить параметры и нажать '$($btn_ok.Text)'"
        }
        else
        {
            [System.Environment]::Exit(1)
        }
    })
    $main_form.Controls.Add($btn_cancel)

    $ProgressBar = New-Object System.Windows.Forms.ProgressBar
#    $ProgressBar.Visible = $false
    $ProgressBar.Location = New-Object System.Drawing.Point(10, 406)
    $ProgressBar.Size     = New-Object System.Drawing.Size(550, 15)    
    $ProgressBar.Value    = 0
    $main_form.Controls.add($ProgressBar)


    $cb_debug = New-Object System.Windows.Forms.CheckBox
    $cb_debug.location = New-Object System.Drawing.Point(10, 430)
    $cb_debug.Size = New-Object System.Drawing.Size(200, 15)    
    $cb_debug.Text = "Отображать полную информацию"
    $cb_debug.TabIndex = 18
    $cb_debug.checkAlign = "MiddleLeft"    
    $cb_debug.Name = "Debug"
    $cb_debug.Checked = $false
    $main_form.Controls.Add($cb_debug) 
    $cb_debug.Add_Click(
    {
        Output-IgnoreOnlyDebug $cb_debug.Checked
    })

    $btn_clear = New-Object System.Windows.Forms.Button
    $btn_clear.Location = New-Object System.Drawing.Point(210, 428)
    $btn_clear.Size = New-Object System.Drawing.Size(78,17)
    $btn_clear.TabIndex = 19
    $btn_clear.Text = 'Очистить'
    $main_form.Controls.Add($btn_clear)
    $btn_clear.Add_Click(
    {
        $tb_output.Text = ""
    })

    $tb_output = New-Object System.Windows.Forms.TextBox
    $tb_output.Location = New-Object System.Drawing.Point(10, 450)
    $tb_output.Size = New-Object System.Drawing.Size(720, 257)    
    $tb_output.Multiline = $true
    $tb_output.Scrollbars = "Vertical"
    $tb_output.ReadOnly = $true
    $main_form.Controls.Add($tb_output)


    # Show window
    $main_form.Add_Shown(
    {
        Output-SetWindow $tb_output
        Output-SetTabСharacter "  "        

        $main_form.ControlBox = $false
        $cb_debug.Enabled     = $false
        $cb_server.Enabled    = $false
        $cb_scada.Enabled     = $false
        $btn_ok.Enabled       = $false
        $btn_cancel.Enabled   = $false
        $btn_clear.Enabled    = $false

        $local:result = SearchInstalledSoftware
        Output-Result $local:result
        if (-not $local:result)
        {
            $btn_cancel.Enabled   = $true
            return
        }

        $ProgressBar.Visible  = $false

        $tb_server_exe.Text   = $script:ServerPath
        $tb_scada_exe.Text    = $script:ScadaPath

        $main_form.ControlBox = $true
        $cb_debug.Enabled     = $true
        $cb_server.Enabled    = $true
        $cb_scada.Enabled     = $true
        $btn_ok.Enabled       = $true
        $btn_cancel.Enabled   = $true
        $btn_clear.Enabled    = $true

        $cb_server.Select()

        Output-Write "● Необходимо заполнить параметры и нажать '$($btn_ok.Text)'" -LinesBefore 1
    })    

    $result = $main_form.ShowDialog()

    if ($result -ne [System.Windows.Forms.DialogResult]::OK)
    {
     #   MessageBox "Процесс настройки прерван, все изменения утеряны!" $global:Defaults.MainWndCaption $ID_OK $ICON_INFO
        Break
    }

    $script:main_form.Dispose()
}

ShowMainForm